# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/core.R' of the Bioconductor DESeq2 package
# (version 3.16).


import logging
import numpy as np
import scipy.stats
from scipy.stats import trim_mean

from .dispersions import estimateDispersionsGeneEst, estimateDispersionsMAP
from .misc import nOrMoreInCell
from .wald import nbinomWaldTest


def refitWithoutOutliers(
    obj,
    test,
    betaPrior,
    full,
    reduced,
    quiet,
    minReplicatesForReplace,
    modelMatrix,
    modelMatrixType,
):
    """ """
    cooks = obj.layers["cooks"]
    obj = replaceOutliers(obj, minReplicates=minReplicatesForReplace)

    # refit without outliers, if there were any replacements
    nrefit = np.nansum(obj.var["replace"])
    if nrefit > 0:
        obj = obj.getBaseMeansAndVariances()
        newAllZero = obj.var["replace"] & obj.var["allZero"]
    # only refit if some of the replacements do not result in all zero counts
    # otherwise, these cases are handled by results()
    if nrefit > 0 and nrefit > np.sum(newAllZero):
        logging.info(
            f"""
        -- replacing outliers and refitting for {nrefit} genes
        -- DESeq argument 'minReplicatesForReplace' = {minReplicatesForReplace}
        -- original counts are preserved in dds.counts()
        """
        )

        # refit on those rows which had replacement
        refitReplace = obj.var["replace"] & ~obj.var["allZero"]
        objSub = obj[:, refitReplace]
        intermediateOrResults = objSub.var.type.filter(
            ["intermediate", "results"]
        ).columns
        objSub.var = objSub.var.drop(intermediateOrResults, axis=1)
        for c in intermediateOrResults:
            del objSub.var.type[c]
            del objSub.var.description[c]

        # estimate gene-wise dispersion
        logging.info("estimating dispersions")
        objSub = estimateDispersionsGeneEst(
            objSub, quiet=quiet, modelMatrix=modelMatrix
        )

        # need to redo fitted dispersion due to changes in base mean
        objSub.var["dispFit"] = obj.dispersionFunction(objSub.var["baseMean"])
        objSub.var.type["dispFit"] = "intermediate"
        objSub.var.description["dispFit"] = "fitted values of dispersion"
        dispPriorVar = obj.dispersionFunction.dispPriorVar

        # estimate dispersion MAP
        objSub = estimateDispersionsMAP(
            objSub, quiet=quiet, dispPriorVar=dispPriorVar, modelMatrix=modelMatrix
        )

        # fit GLM
        logging.info("fitting model and testing")
        if test == "Wald":
            betaPriorVar = obj.betaPriorVar
            objSub = nbinomWaldTest(
                objSub,
                betaPrior=betaPrior,
                betaPriorVar=betaPriorVar,
                quiet=quiet,
                modelMatrix=modelMatrix,
                modelMatrixType=modelMatrixType,
            )
        elif test == "LRT":
            objSub = nbinomLRT(objSub, full=full, reduced=reduced, quiet=quiet)

        obj.var.loc[refitReplace, objSub.var.columns] = objSub.var
        obj.var.loc[newAllZero, obj.var.type.filter("results").columns] = np.nan

        # continue to flag if some conditions have less than minReplicatesForReplace
        if np.all(obj.obs["replaceable"]):
            obj.var["maxCooks"] = np.nan
        else:
            replaceCooks = obj.layers["cooks"].copy()
            replaceCooks[obj.obs["replaceable"]] = 0
            obj.var["maxCooks"] = recordMaxCooks(
                obj.design, obj.obs, obj.dispModelMatrix, replaceCooks, obj.n_vars
            )

    if nrefit > 0:
        # save the counts used for fitting as replaceCounts
        obj.layers["replaceCounts"] = obj.counts().copy()
        obj.layers["replaceCooks"] = obj.layers["cooks"].copy()

        # preserve original counts and Cook's distances
        obj.X = obj.layers["originalCounts"]
        obj.layers["cooks"] = cooks

        # no longer needed
        del obj.layers["originalCounts"]

    return obj


def replaceOutliers(
    obj, trim=0.2, cooksCutoff=None, minReplicates=7, whichSamples=None
):
    """
    Replace outliers with trimmed mean

    Note that this function is called within :func:`DESeq`, so is not necessary
    to call on top of a :func:`DESeq` call. See the documentation for
    :code:`minReplicatesForReplace` in :func:`DESeq`.

    This function replaces outlier counts flagged by extreme Cook's distances,
    as calculated by :func:`DESeq`, :func:`nbinomWaldTest` or
    :func:`nbinomLRT`, with values predicted by the trimmed mean over all
    samples (and adjusted by size factor or normalization factor).  This
    function replaces the counts in the matrix returned by :code:`dds.counts()`
    and the Cook's distances in :code:`dds.layers["cook"]`. Original counts are
    preserved in :code:`dds.layers["originalCounts"]`.

    The :func:`DESeq` function calculates a diagnostic measure called Cook's
    distance for every gene and every sample. The :meth:`.DESeqDataSet.results`
    function then sets the p-values to NA for genes which contain an outlying
    count as defined by a Cook's distance above a threshold. With may degrees
    of freedom, i.e. many more samples than number of parameters to be
    estimated, it might be undesirable to remove entire genes fomr the analysis
    just because their data include a single count outlier.  An alternative
    strategy is to replace the outlier counts with the trimmed mean over all
    samples, adjusted by the size factor or normalization factor for that
    sample. The following simple function performs this replacement for the
    user, for samples which have at least :code:`minReplicates` number of
    replicates (including that sample). For more information on Cook's
    distance, please see the two sections of the module documentation: "Dealing
    with count outliers" and "Count outlier detection".

    Arguments
    ---------
    obj : DESeqDataSet
        a DESeqDataSet that has already been processed by either :func:`DESeq`,
        :func:`nbinomWaldTest` or :func:`nbinomLRT`, and therefore contains a
        matrix of Cook's distances (used to define the outlier counts) in
        :code:`obj.layers["cooks"]`.
    trim : float
        the fraction (0 to 0.5) of observations to be trimmed from each end of
        the normalized counts for a gene before the mean is computed.
    cooksCutoff : float
        the threshold for defining an outlier to be replaced. Defaults to the
        .99 quantile of the :math:`F(p, m-p)` distribution, where :math:`p` is
        the number of parameters and :math:`m` is the number of samples.
    minReplicates : int
        the minimum number of replicate samples necessary to consider a sample
        eligible for replacement (including itself). Outlier counts will not be
        replaced if the sample is in a cell which has less than
        :code:`minReplicates` replicates.
    whichSamples : array-like, optional
        a numeric or logical index to specify which samples should have
        outliers replaced. If missing, this is determined using
        :code:`minReplicates`.

    Return
    ------
    DESeqDataSet
        the input :code:`obj` with replaced counts in the slot returned by
        :meth:`.DESeqDataSet.counts`, and the original counts preserved in
        :code:`obj.layers["originalCounts"]`.
    """
    if obj.modelMatrix is None or "cooks" not in obj.layers:
        raise ValueError(
            "first run DESeq, nbinomWaldTest or nbinomLRT to identify outliers"
        )
    if not isinstance(minReplicates, int):
        raise ValueError(f"invalid value for minReplicates: {minReplicates}")
    if minReplicates < 3:
        raise ValueError(
            "at least 3 replicates are necessary in order to identify a sample as a count outlier."
        )

    p = obj.modelMatrix.shape[1]
    m = obj.n_obs
    if m <= p:
        obj.layers["originalCounts"] = obj.counts().copy()
        return obj
    if cooksCutoff is None:
        cooksCutoff = scipy.stats.f.ppf(0.99, p, m - p)

    idx = obj.layers["cooks"] > cooksCutoff
    obj.var["replace"] = np.any(obj.layers["cooks"] > cooksCutoff, axis=0)
    obj.var.type["replace"] = "intermediate"
    obj.var.description["replace"] = "had counts replaced"

    trimBaseMean = trim_mean(obj.counts(normalized=True), trim)
    # build a matrix of counts based on the trimmed mean and the size factors
    if obj.normalizationFactors is not None:
        replacementCounts = (
            np.repeat(trimBaseMean[None], obj.n_obs, axis=0)
            * obj.normalizationFactors.values[:, None]
        )
    else:
        replacementCounts = (
            np.repeat(trimBaseMean[None], obj.n_obs, axis=0)
            * obj.sizeFactors.values[:, None]
        )

    # replace only those values which fall above the cutoff on Cook's distance
    newCounts = obj.counts().copy()
    newCounts[idx] = replacementCounts[idx]

    if whichSamples is None:
        whichSamples = nOrMoreInCell(obj.modelMatrix, n=minReplicates)

    whichSamples.index = obj.obs_names
    obj.obs["replaceable"] = whichSamples
    obj.obs.type["replaceable"] = "intermediate"
    obj.obs.description["replaceable"] = "outliers can be replaced"
    obj.layers["originalCounts"] = obj.counts().copy()
    obj.X[whichSamples] = newCounts[whichSamples]
    return obj



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/adjustedProfileLik.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np
from .makeCompressedMatrix import (
    _compressOffsets,
    _compressDispersions,
    _compressWeights,
)
from .glmFit import glmFit
from .edgepy_cpp import cxx_compute_apl


def adjustedProfileLik(
    dispersion, y, design, offset, weights=None, adjust=True, start=None, get_coef=False
):
    """
    Compute adjusted profile log-likelihoods for the dispersion parameters of
    genewise negative binomial GLMs.

    For each row of data, compute the adjusted profile log-likelihood for the
    dispersion parameter of the negative binomial GLM. The adjusted log
    likelihood is described by McCarthy et al. (2012) [2]_ and is based on the
    method of Cox and Reid (1987) [1]_.

    The adjusted profile likelihood is an approximation to the log-likelihood
    function, conditional on the estimated values of the coefficients in the NB
    log-linear models. The conditions likelihood approach is a technique for
    adjusting the likelihood function to allow for the fact that nuisance
    parameters have to be estimated in order to evaluate the likelihood. When
    estimating the dispersion, the nuisance parameters are the coefficients in
    the log-linear models.

    This implementation calls the LAPACK library to perform the Cholesky
    decomposition during adjustment estimation.

    The purpose of :code:`start` and :code:`get_coef` is to allow hot-starting
    for multiple calls to `adjustedProfileLik`, when only :code:`dispersion` is
    altered.  Specifically, the returned GLM coefficients from one call with
    :code:`get_coef=True` can be used as the :code:`start` values for the next
    call.

    The :code:`weights` argument is interpreted in terms of averages. Each
    value of :code:`y` is assumed to be the average of :code:`n` i.i.d NB
    counts, where :code:`n` is given by the weight. This assumption can be
    generalized to fractional weights.

    Arguments
    ---------
    dispersion : float or array
        vector of dispersions
    y : matrix
        matrix of counts
    design : matrix
        design matrix
    offset : matrix or vector or float
        matrix of same shape as :code:`y` giving offsets for the log-linear
        models.  Can also be scalar or a vector of length :code:`y.shape[1]`,
        in which case it is broadcasted to the same shape as :code:`y`.
    weights : matrix, optional
        numeric matrix giving observation weights
    adjust : bool, optional
        if `True` then Cox-Reid adjustment is made to the log-likelihood.
        if `False` then the log-likelihood is returned without adjustment.
    start : matrix, optional
        numeric matrix of starting values for the GLM coefficients, to be
        passed to :func:`glmFit`.
    get_coef : bool, optional
        specifying whether fitted GLM coefficients should be returned

    Returns
    -------
    array
        vector of adjusted profile log-likelihood values is returned containing
        one element for each row of `y`.
    matrix (only if `get_coef` is `True`)
        the matrix of fitted GLM coefficients

    References
    ----------
    .. [1] D. R. Cox, N. Reid. 1987. Parameter orthogonality and approximate
       conditional inference. Journal of the Royal Statistical Society Series B
       49, 1-39.
    .. [2] D. J. McCarthy, Y. Chen, G. K. Smyth. 2012. Differential expression
       analysis of multifactor RNA-Seq experiments with respect to biological
       variation. Nucleic Acids Research 40, 4288-4297. :doi:`10.1093/nar/gks042`
    """

    # Checking counts
    y = np.asarray(y, order="F")

    # Checking offsets
    offset = _compressOffsets(y, offset=offset)

    # Checking dispersion
    dispersion = _compressDispersions(y, dispersion)

    # Checking weights
    weights = _compressWeights(y, weights)

    # Fit tagwise linear models
    fit = glmFit(
        y,
        design=design,
        dispersion=dispersion,
        offset=offset,
        prior_count=0,
        weights=weights,
        start=start,
    )
    mu = fit.fitted_values
    assert mu.dtype == np.dtype("double")
    assert mu.flags.f_contiguous

    # Compute adjusted log-likelihood
    apl = cxx_compute_apl(y, mu, dispersion, weights, adjust, design)

    # Deciding what to return
    if get_coef:
        return (apl, fit.coefficients)
    else:
        return apl



import unittest
import numpy as np
import pandas as pd

from inmoose.utils import Factor
from inmoose.deseq2 import makeExampleDESeqDataSet, DESeq, nbinomWaldTest
from inmoose.deseq2.fitNbinomGLMs import fitNbinomGLMsOptim


class Test(unittest.TestCase):
    def test_weights(self):
        """test that weights work"""
        dds = makeExampleDESeqDataSet(n=10, seed=42)
        w = np.ones(dds.shape)
        w[0, 0] = 0
        # check that weight to 0 is like a remove sample
        dds = DESeq(dds, quiet=True)
        dds2 = dds.copy()
        dds2.layers["weights"] = w
        dds2 = nbinomWaldTest(dds2)
        dds3 = dds.copy()[1:, :]
        dds3 = nbinomWaldTest(dds3)

        # in terms of LFC, SE and deviance
        self.assertAlmostEqual(
            dds2.results().log2FoldChange[0],
            dds3.results().log2FoldChange[0],
            delta=1e-5,
        )
        self.assertAlmostEqual(
            dds2.results().lfcSE[0], dds3.results().lfcSE[0], delta=1e-6
        )
        self.assertAlmostEqual(
            dds2.var["deviance"][0], dds3.var["deviance"][0], delta=1e-7
        )

        # check weights working in the optim code
        nf = np.repeat(dds.sizeFactors.values[:, None], dds.n_vars, axis=1)
        o = fitNbinomGLMsOptim(
            obj=dds,
            modelMatrix=dds.design,
            lambda_=np.repeat(1e-6, 2),
            colsForOptim=[0],
            colStable=np.repeat(True, dds.n_vars),
            normalizationFactors=nf,
            alpha_hat=dds.var["dispersion"],
            weights=w,
            useWeights=True,
            betaMatrix=pd.DataFrame(np.zeros((dds.n_vars, 2))),
            betaSE=pd.DataFrame(np.zeros((dds.n_vars, 2))),
            betaConv=np.repeat(False, dds.n_vars),
            beta_mat=np.zeros((dds.n_vars, 2)),
            mu=np.zeros(dds.shape),
            logLike=np.zeros(dds.n_vars),
        )
        self.assertAlmostEqual(
            dds3.results().log2FoldChange[0], o["betaMatrix"].iloc[0, 1], delta=1e-5
        )

    def test_weights_with_beta_prior(self):
        """test that weights can be used with betaPrior=True"""
        dds = makeExampleDESeqDataSet(n=10, seed=42)
        w = np.ones(dds.shape)
        w[0, 0] = 0
        dds.layers["weights"] = w
        dds = DESeq(dds, betaPrior=True, quiet=True)

        # check weights working for intercept only
        dds.design = "~1"
        dds = DESeq(dds, quiet=True)
        dds2 = dds
        dds2.layers["weights"] = w
        dds2 = nbinomWaldTest(dds2)
        dds3 = dds[1:, :]
        dds3 = nbinomWaldTest(dds3)

        self.assertEqual(
            dds2.results().log2FoldChange[0], dds3.results().log2FoldChange[0]
        )
        self.assertEqual(dds2.results().lfcSE[0], dds3.results().lfcSE[0])
        self.assertEqual(dds2.var["deviance"][0], dds3.var["deviance"][0])

    def test_weights_downweight_outliers(self):
        """test that weights downweight outlier in dispersion estimation"""
        dds = makeExampleDESeqDataSet(n=10, seed=42)
        dds.counts()[0, 0] = 100
        dds.sizeFactors = np.ones(dds.n_obs)
        dds = dds.estimateDispersions()
        dds2 = dds.copy()
        w = np.ones(dds.shape)
        w[0, 0] = 0
        dds2.layers["weights"] = w
        dds2 = dds2.estimateDispersions()
        dds3 = dds.copy()[1:, :]
        dds3 = dds3.estimateDispersions()

        self.assertAlmostEqual(
            dds2.var["dispGeneEst"][0], dds3.var["dispGeneEst"][0], delta=1e-1
        )
        # MAP estimates won't be equal because of different dispersion prior widths
        self.assertGreater(dds.var["dispMAP"][0], dds2.var["dispMAP"][0])

    def test_weights_warning(self):
        """test that weights failing check gives warning, passes them through"""
        dds = makeExampleDESeqDataSet(n=10, seed=42)
        w = np.ones(dds.shape)
        w[0:6, 0] = 0
        dds.layers["weights"] = w
        with self.assertWarnsRegex(
            UserWarning,
            expected_regex="for 1 genes, the weights as supplied won't allow parameter estimation",
        ):
            dds = DESeq(dds)
        self.assertTrue(dds.var["allZero"][0])
        self.assertTrue(dds.var["weightsFail"][0])
        res = dds.results()

    @unittest.skip("not sure what is tested here")
    def test_weights_CR(self):
        """test that weights with and without CR term included"""
        alpha = 0.25

        def dmr(x):
            return alpha

        dds = makeExampleDESeqDataSet(
            n=50, m=100, betaSD=1, interceptMean=10, interceptSD=0.5, dispMeanRel=dmr
        )
        dds.obs["group"] = Factor(np.repeat(np.arange(50), 2))
        dds.design = "~0 + group + condition"
        w = np.ones(dds.shape)
        o = 35
        w[0:o, :] = 1e-6
        w[50 : (50 + o), :] = 1e-6
        dds.layers["weights"] = w
        dds.counts()[1:o, :] = 1
        dds.counts()[50 : (50 + o), :] = 1
        dds.sizeFactors = 1
        dds = dds.estimateDispersions(fitType="mean")
        dds2 = dds.estimateDispersions(fitType="mean", useCR=False)



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/vst.R' Bioconductor DESeq2 package (version
# 3.16).


from anndata import AnnData
import numpy as np

from . import (
    DESeqDataSet,
    DESeqTransform,
    estimateDispersionsGeneEst,
    estimateDispersionsFit,
)


def varianceStabilizingTransformation(obj, blind=True, fitType="parametric"):
    """
    Apply a variance stabilizing transformation (VST) to the count data

    This function calculates a variance stabilizing transformation (VST) from
    the fitted dispersion-mean relation(s) and then transforms the count data
    (normalized by division by the size factors or normalization factors),
    yielding a matrix of values which are now approximately homoskedastic
    (having constant variance along the range of mean values). The
    transformation also normalizes with respect to library size. The
    :func:`rlog` is less sensitive to size factors, which can be an issue when
    size factors vary widely. These transformations are useful when checking
    for outliers or as input for machine learning techniques such as clustering
    or linear discriminant analysis.

    Variance stabilizing transformation was originally described in [Anders2010]_.

    Details
    -------
    For each sample (*i.e.* line of :code:`dds.counts()`), the full variance
    function is calculated from the raw variance (by scaling according to the
    size factor and adding the shot noise). We recommend a blind estimation of
    the variance function, *i.e.* one ignoring conditions. This is performed by
    default, and can be modified using the :code:`blind` argument.

    Note that neither :func:`rlog` transformation nor the VST are used by the
    differential expression estimation in :func:`DESeq`, which always occurs on
    the raw count data, through generalized linear modeling which incorporates
    knowledge of the variance-mean dependence. The :func:`rlog` transformation
    and VST are offered as separate functionality which can be used for
    visualization, clustering or other machine learning tasks. See the
    transformation section of the vignette for more details.

    The transformation does not require that one has already estimated size
    factors and dispersions.

    A typical workflow is shown in Section *Variance stabilizing
    transformation* in the vignette.

    If :func:`estimateDispersions` was called with:

    - :code:`fitType="parametric"`: a closed-form expression for the variance
      stabilizing transformation is used on the normalized count data.
    - :code:`fitType="local"`: the reciprocal of the square root of the
      variance of the normalized counts, as derived from the dispersion fit, is
      then numerically integrated, and the integral (approximated by a spline
      function) is evaluated for each count value in the column, yielding a
      transformed value.
    - :code:`fitType="mean"`, a VST is applied for Negative Binomial
      distributed counts, :math:`k`, with a fixed dispersion, :math:`a`:
      :math:`(2 \\operatorname{asinh}(\sqrt{a k}) - \log(a) - \log(4)) / \log(2)`.

    In all cases, the transformation is scaled such that for large counts, it
    becomes asymptotically (for large values) equal to the logarithm to base 2
    of normalized counts.

    The variance stabilizing transformation from a previous dataset can be
    "frozen" and reapplied to new samples. The frozen VST is accomplished by
    saving the dispersion function accessible with :func:`dispersionFunction`,
    assigning this to the :class:`DESeqDataSet` with the new samples, and
    running :func:`varianceStabilizingTransformation` with :code:`blind` set to
    :code:`False`. Then the dispersion function from the previous dataset will
    be used to transform the new sample(s).

    Limitations: In order to preserve normalization, the same transformation
    has to be used for all samples. This results in the variance stabilization
    to be only approximate. The more the size factors differ, the more residual
    dependence of the variance on the mean will be found in the transformed
    data. :func:`rlog` is a transformation which can perform better in these
    cases. As shown in the vignette, :code:`meanSdPlot` from the package *vsn*
    can be used to see whether this is a problem.

    Arguments
    ---------
    obj : DESeqDataSet or matrix
        a :class:`DESeqDataSet` or matrix of counts
    blind : bool
        whether to blind the transformation to the experimental design.
        :code:`blind=True` should be used for comparing samples in a manner
        unbiased by prior information on samples, for example to perform sample
        QA (quality assurance).
        :code:`blind=False` should be used for transforming data for downstream
        analysis, where the full use of the design information should be made.
        :code:`blind=False` will skip re-estimation of the dispersion trend, if
        this has already been calculated.
        If many genes have large differences in counts due to the experimental design,
        it is important to set :code:`blind=False` for downstream analysis.
        Defaults to :code:`True`.
    fitType : { "parametric", "local", "mean" }
        in case dispersions have not yet been estimated for :code:`self`, this
        parameter is passed on to :func:`estimateDispersions` (options described
        there).
        Defaults to :code:`"parametric"`.

    Returns
    -------
    DESeqTransform or matrix
        returns a :class:`DESeqTransform` if a :class:`DESeqDataSet` was
        provided, or returns a matrix if a count matrix was provided.
        Note that for :class:`DESeqTransform` output, the matrix of transformed
        values is stored in :code:`vsd.layers`.
    """
    if not isinstance(obj, DESeqDataSet):
        matrixIn = True
        obj = DESeqDataSet(obj, design="~1")
    else:
        obj = obj.copy()
        matrixIn = False

    if obj.sizeFactors is None and obj.normalizationFactors is None:
        obj = obj.estimateSizeFactors()

    if blind:
        obj.design = "~1"

    if blind or obj.dispersionFunction.fitType is None:
        obj = estimateDispersionsGeneEst(obj, quiet=True)
        obj = estimateDispersionsFit(obj, fitType, quiet=True)

    vsd = getVarianceStabilizedData(obj)
    if matrixIn:
        return vsd

    # TODO missing metadata
    ad = AnnData(vsd, obs=obj.obs, var=obj.var)
    return DESeqTransform(ad)


def getVarianceStabilizedData(obj):
    if obj.dispersionFunction.fitType is None:
        raise ValueError(
            "call estimateDispersions before calling getVarianceStabilizedData"
        )

    ncounts = obj.counts(normalized=True)
    if obj.dispersionFunction.fitType == "parametric":
        coefs = obj.dispersionFunction.coefficients
        vst_fn = lambda q: np.log(
            (
                1
                + coefs["extraPois"]
                + 2 * coefs["asymptDisp"] * q
                + 2
                * np.sqrt(
                    coefs["asymptDisp"]
                    * q
                    * (1 + coefs["extraPois"] + coefs["asymptDisp"] * q)
                )
            )
            / (4 * coefs["asymptDisp"])
        ) / np.log(2)
        return vst_fn(ncounts)

    elif obj.dispersionFunction.fitType == "local":
        # non-parametric fit -> numerical integration
        if obj.sizeFactors is None:
            if obj.normalizationFactors is None:
                raise ValueError(
                    "both sizeFactors and normalizationFactors are missing!"
                )
            sf = np.exp(np.log(obj.normalizationFactors).mean(axis=1))
        else:
            sf = obj.sizeFactors
        xg = np.sinh(np.linspace(np.arcsinh(0), np.arcsinh(ncounts.max()), num=1000))[
            1:
        ]
        xim = (1 / sf).mean()
        baseVarsAtGrid = obj.dispersionFunction(xg) * xg**2 + xim * xg
        integrand = 1 / np.sqrt(baseVarsAtGrid)
        splf = splinefun(
            np.arcsinh((xg[1:] + xg[:-1]) / 2),
            ((xg[1:] - xg[:-1]) * (integrand[1:] + integrand[:-1]) / 2).cumsum(),
        )
        h1 = quantile(ncounts.mean(axis=1), 0.95)
        h2 = quantile(ncounts.mean(axis=1), 0.999)
        eta = (np.log2(h2) - np.log2(h1)) / (
            splf(np.arcsinh(h2)) - splf(np.arcsinh(h1))
        )
        xi = np.log2(h1) - eta * splf(np.arcsinh(h1))
        tc = [
            eta * splf(np.arcsinh(ncounts[:, clm])) + xi
            for clm in obj.counts().var_names
        ]
        tc.rownames = obj.counts().rownames
        return tc
    elif obj.dispersionFunction.fitType == "mean":
        alpha = obj.dispersionFunction.mean
        # the following stabilizes NB counts with fixed dispersion alpha
        # and converges to log2(q) as q => infinity
        vst_fn = lambda q: (
            2 * np.arcsinh(np.sqrt(alpha * q)) - np.log(alpha) - np.log(4)
        ) / np.log(2)
        return vst_fn(ncounts)
    else:
        raise ValueError("fitType is not parametric, local or mean")


def vst(obj, blind=True, nsub=1000, fitType="parametric"):
    """
    Quickly estimate dispersion trend and apply a variance stabilizing transformation

    This is a wrapper for the :func:`varianceStabilizingTransformation` (VST)
    that provides much faster estimation of the dispersion trend used to
    determine the formula for the VST. The speed-up is accomplished by
    subsetting to a smaller number of genes in order to estimate this
    dispersion trend. The subset of genes is chosen deterministically, to span
    the range of genes' mean normalized counts.

    Arguments
    ---------
    obj : DESeqDataSet or matrix
        a :class:`DESeqDataSet` or matrix of counts
    blind : bool
        whether to blind the transformation to the experimental design (see
        :func:`varianceStabilizingTransformation`)
    nsub : int
        the number of genes to subset to (default 1000)
    fitType : { "parametric", "local", "mean" }
        for estimation of dispersions: this parameter is passed on to
        :func:`estimateDispersions` (options described there)

    Returns
    -------
    DESeqTransform or matrix
        same as :func:`varianceStabilizingTransformation`
    """
    if obj.n_var < nsub:
        raise ValueError(
            f"Object has less than {nsub} rows, it is recommended to use varianceStabilizingTransformation directly"
        )

    if not isinstance(obj, DESeqDataSet):
        matrixIn = True
        obj = DESeqDataSet(obj, design="~1")
    else:
        if blind:
            obj.design = "~1"
        matrixIn = False

    if obj.sizeFactors is None and obj.normalizationFactors is None:
        obj = obj.estimateSizeFactors()
    baseMean = obj.counts(normalized=True).mean(axis=0)
    if (baseMean > 5).sum() < nsub:
        raise ValueError(
            "Object has less than {nsub} genes with mean normalized count > 5, it is recommended to use varianceStabilizingTransformation directly."
        )

    # subset to a specified number of genes with mean normalized count > 5
    obj_sub = obj[:, baseMean > 5]
    baseMean = baseMean[baseMean > 5]
    o = np.argsort(baseMean)
    idx = o[np.linspace(0, len(o) - 1, num=nsub, dtype=int)]
    obj_sub = obj_sub[idx, :]

    # estimate dispersion trend
    obj_sub = estimateDispersionsGeneEst(obj_sub, quiet=True)
    obj_sub = estimateDispersionsFit(obj_sub, fitType=fitType, quiet=True)

    # assign to the full object
    obj.setDispFunction(obj_sub.dispersionFunction)

    # calculate and apply the VST (note blinding is accomplished above, here
    # :code:`blind=False` is used to avoid re-calculating dispersion)
    vsd = varianceStabilizingTransformation(obj, blind=False)
    if matrixIn:
        return vsd.X
    else:
        return vsd



import unittest
import numpy as np

from inmoose.utils import Factor
from inmoose.deseq2 import DESeq, makeExampleDESeqDataSet


class zero_zero(unittest.TestCase):
    def test_zero_zero(self):
        """test that contrast of two groups with all zeros - LFC zeroed out"""

        # test comparison of two groups with all zeros
        dds = makeExampleDESeqDataSet(
            m=8, n=100, sizeFactors=[1, 1, 0.5, 0.5, 1, 1, 2, 2]
        )
        dds.obs["condition"] = Factor(np.repeat(["A", "B", "C", "D"], 2))
        dds.design = "~condition"
        dds.counts()[:, 0] = [100, 110, 0, 0, 100, 110, 0, 0]
        dds.counts()[:, 1] = np.repeat(0, 8)
        dds = DESeq(dds)

        res = dds.results(contrast=["condition", "D", "B"])
        self.assertEqual(res.log2FoldChange[0], 0)
        res = dds.results(contrast=[0, -1, 0, 1])
        self.assertEqual(res.log2FoldChange[0], 0)
        res = dds.results(contrast=["condition_D_vs_A", "condition_B_vs_A"])
        self.assertEqual(res.log2FoldChange[0], 0)

        res = dds.results(name="condition_D_vs_A")
        self.assertNotEqual(res.log2FoldChange[0], 0)
        res = dds.results([0, 0, 0, 1])
        self.assertNotEqual(res.log2FoldChange[0], 0)

        # if all samples have 0, should be NA
        res = dds.results(contrast=["condition", "D", "B"])
        self.assertTrue(np.isnan(res.log2FoldChange[1]))
        res = dds.results(contrast=[0, -1, 0, 1])
        self.assertTrue(np.isnan(res.log2FoldChange[1]))



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/makeCompressedMatrix.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np


# NB: we do not fully implement edgeR class CompressedMatrix: we use fully-expanded 2D numpy arrays instead
#     still, to avoid useless calls to `_compress*` and `check_finite`, we subclass ndarray
#     note that the name `CompressedMatrix` for the subclass is only historical
class CompressedMatrix(np.ndarray):
    def __new__(cls, input_array):
        # Input array is an already formed ndarray instance
        # We thus cast to be our class type
        return np.asarray(input_array, order="F").view(cls)


def makeCompressedMatrix(x, dims, byrow=True):
    """
    Coerce a None, scalar, vector or matrix to a compressed matrix.
    """
    if x.__class__ == CompressedMatrix:
        return x

    if len(dims) != 2:
        raise ValueError("dims does not represent the shape of a matrix")
    x = np.asarray(x, order="F")
    if len(x.shape) == 0:
        x = np.full(dims, x, order="F")
    elif len(x.shape) == 2:
        assert dims == x.shape
    elif len(x.shape) == 1:
        if not byrow:
            if dims[0] != len(x):
                raise ValueError("dims[0] should be equal to length of x")
            # build matrix by duplicating the column x
            x = np.array([x for i in range(dims[1])], order="C").T
            # transposing a C-array should yield a F-array, but better safe than sorry
            x = np.asarray(x, order="F")
        else:
            if dims[1] != len(x):
                raise ValueError("dims[1] should be equal to length of x")
            # build matrix by duplicating the row x
            x = np.array([x for i in range(dims[0])], order="F")
    else:
        raise ValueError("input has too many dimensions to be interpreted as a matrix")

    return CompressedMatrix(x)


def check_finite(x, what, negative_allowed):
    xmin = np.amin(x)
    if np.isnan(xmin):
        raise ValueError("NaN " + what + " not allowed")
    if not negative_allowed and xmin < 0:
        raise ValueError("negative " + what + " not allowed")
    if np.isinf(np.nanmax(x)):
        raise ValueError("infinite " + what + " is not allowed")


def _compressOffsets(y, offset, lib_size=None):
    """
    Check for finite values
    If provided, offset takes precedence over lib_size
    If neither are provided, lib_size is automatically as the sum of counts in the count matrix y
    """
    if offset.__class__ == CompressedMatrix:
        return offset

    if offset is None:
        if lib_size is None:
            lib_size = y.sum(axis=0)
        offset = np.log(lib_size)

    offset = np.asarray(offset, order="F", dtype="double")
    offset = makeCompressedMatrix(offset, y.shape, byrow=True)
    check_finite(offset, "offset", negative_allowed=True)
    return offset


def _compressWeights(y, weights=None):
    """
    Check for non-negative finite values
    All weights default to 1 when not specified
    """
    if weights.__class__ == CompressedMatrix:
        return weights

    if weights is None:
        weights = 1

    weights = np.asarray(weights, order="F", dtype="double")
    weights = makeCompressedMatrix(weights, y.shape, byrow=True)
    check_finite(weights, "weights", negative_allowed=False)
    return weights


def _compressDispersions(y, dispersion):
    """
    Check for non-negative finite values
    """
    if dispersion.__class__ == CompressedMatrix:
        return dispersion

    dispersion = np.asarray(dispersion, order="F", dtype="double")
    dispersion = makeCompressedMatrix(dispersion, y.shape, byrow=False)
    check_finite(dispersion, "dispersion", negative_allowed=False)
    return dispersion


def _compressPrior(y, prior_count):
    """
    Check for non-negative finite values
    """
    if prior_count.__class__ == CompressedMatrix:
        return prior_count

    prior_count = np.asarray(prior_count, order="F", dtype="double")
    prior_count = makeCompressedMatrix(prior_count, y.shape, byrow=False)
    check_finite(prior_count, "prior counts", negative_allowed=False)
    return prior_count



import unittest
import numpy as np
from scipy.stats import f

from inmoose.deseq2 import makeExampleDESeqDataSet, DESeq, replaceOutliers


class Test(unittest.TestCase):
    def test_outlier_filtering_replacement(self):
        """test that outlier filtering and replacement works as expected"""
        dds = makeExampleDESeqDataSet(
            n=100, m=12, dispMeanRel=lambda x: 4 / x + 0.5, seed=42
        )
        dds.X[:, 0] = 0
        dds.X[:, 1] = 10
        dds.X[0, 1] = 100000
        dds.X[:, 2] = 0
        dds.X[0, 2] = 100000

        dds0 = DESeq(dds.copy(), minReplicatesForReplace=np.inf)
        dds1 = DESeq(dds.copy(), minReplicatesForReplace=6)
        pval0 = dds0.results().pvalue[0:3]
        pval1 = dds1.results().pvalue[0:3]
        LFC0 = dds0.results().log2FoldChange[0:3]
        LFC1 = dds1.results().log2FoldChange[0:3]

        # filtered
        self.assertTrue(np.all(np.isnan(pval0)))
        # not filtered
        self.assertTrue(np.all(~np.isnan(pval1[1:3])))
        # counts still the same
        self.assertTrue(np.all(dds1.counts() == dds.counts()))
        # first is NA
        self.assertTrue(np.isnan(LFC1[0]))
        # replaced, reduced LFC
        self.assertTrue(np.abs(LFC1[1]) < np.abs(LFC0[1]))
        # replaced, LFC now zero
        self.assertTrue(LFC1[2] == 0)
        idx = ~dds1.var["replace"]
        # the pvalue for those not replaced is equal
        self.assertTrue(
            np.array_equal(
                dds1.results().pvalue[idx], dds0.results().pvalue[idx], equal_nan=True
            )
        )

        # check that outlier filtering catched throughout range of mu
        beta0 = np.linspace(1, 16, 100)
        idx = np.repeat(np.repeat([True, False], [1, 9]), 10)
        for disp0 in [0.01, 0.1]:
            for m in [10, 20, 80]:
                dds = makeExampleDESeqDataSet(
                    n=100,
                    m=m,
                    interceptMean=beta0,
                    interceptSD=0,
                    dispMeanRel=lambda x: disp0,
                    seed=42,
                )
                dds.counts()[0, idx] = 1000 * 2 ** beta0[idx]
                dds = DESeq(
                    dds, minReplicatesForReplace=np.inf, quiet=True, fitType="mean"
                )
                res = dds.results()
                cutoff = f.ppf(0.99, 2, m - 2)
                outlierCooks = dds.layers["cooks"][0, idx] > cutoff
                nonoutlierCooks = dds.var["maxCooks"][~idx] < cutoff
                self.assertTrue(np.all(np.isnan(res.pvalue[idx])))
                self.assertTrue(np.all(outlierCooks))
                self.assertTrue(np.all(nonoutlierCooks))

        # TODO LRT not implemented so far
        # dds = makeExampleDESeqDataSet(n=100)
        # dds.counts()[0,0] = 1000000
        # dds = DESeq(dds, test="LRT", reduced="~1", minReplicatesForReplace=6)

        # test replace function
        dds = makeExampleDESeqDataSet(n=100, m=4)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="first run DESeq, nbinomWaldTest or nbinomLRT to identify outliers",
        ):
            replaceOutliers(dds)
        dds = DESeq(dds)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="at least 3 replicates are necessary in order to identify a sample as a count outlier",
        ):
            replaceOutliers(dds, minReplicates=2)

        # check model matrix standard bug
        dds = makeExampleDESeqDataSet(n=100, m=20)
        dds.counts()[:, 0] = 0
        dds.counts()[0, 0] = 100000
        dds = DESeq(dds, modelMatrixType="standard")

    def test_outlier_filtering_small_counts(self):
        """test that outlier filtering does not flag small counts"""
        dds = makeExampleDESeqDataSet(n=100, m=8, dispMeanRel=lambda x: 0.01, seed=42)
        dds.counts()[:, 0] = [0, 0, 0, 100, 2100, 2200, 2300, 2400]
        dds.counts()[0, 1:2] = 100000
        dds.counts()[0, 3] = 0
        dds = DESeq(dds, fitType="mean")
        res = dds.results()
        self.assertFalse(np.isnan(res.pvalue[0]))
        self.assertTrue(np.all(np.isnan(res.pvalue[1:2])))



import unittest

from inmoose.data.airway import airway


class Test(unittest.TestCase):
    def test_airway(self):
        """
        check that the airway dataset is properly loaded
        """
        adata = airway()
        self.assertEqual(adata.shape, (8, 64102))
        self.assertEqual(len(adata.obs.columns), 9)



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/maximizeInterpolant.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np

from .edgepy_cpp import cxx_maximize_interpolant


def maximizeInterpolant(x, y):
    """
    Maximize a function given a table of values by spline interpolation.

    Calculate the cubic spline interpolant for each row with the method of
    Forsythe et al. (1977) [1]_, then calculate the derivatives of the spline
    segments adjacent to the input with the maximum function value. This allows
    identification of the maximum of the interpolating spline.

    Arguments
    ---------
    x : array_like
        vector of inputs for the function
    y : array_like
        matrix of function values at the values of :code:`x`. Columns correspond
        to :code:`x` values and each row corresponds to a different function to
        be maximized.

    Returns
    -------
    ndarray
        vector of input values at which the function maxima occur

    References
    ----------
    .. [1] G. E. Forsythe, M. A. Malcolm, C. B. Moler. 1977. Computer Methods
       for Mathematical Computations. Prentice-Hall.
    """
    x = np.asarray(x, order="F", dtype="double")
    y = np.asarray(y, order="F", dtype="double")
    if len(y.shape) != 2:
        raise ValueError("y is not a matrix: cannot perform interpolation")
    if len(x) != y.shape[1]:
        raise ValueError("number of columns must equal number of spline points")
    if not np.array_equal(np.unique(x), x):
        raise ValueError("spline points must be unique and sorted")

    # Performing some type checking
    out = cxx_maximize_interpolant(x, y)
    return out



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/core.R' of the Bioconductor DESeq2 package
# (version 3.16).


import logging
import numpy as np
import pandas as pd
import patsy

from .DESeqDataSet import DESeqDataSet, checkFullRank
from .lrt import checkLRT, nbinomLRT
from .misc import nOrMoreInCell
from .outliers import refitWithoutOutliers
from .wald import nbinomWaldTest


def DESeq(
    obj,
    test="Wald",
    fitType="parametric",
    sfType="ratio",
    betaPrior=False,
    full=None,
    reduced=None,
    quiet=False,
    minReplicatesForReplace=7,
    modelMatrixType=None,
    useT=False,
    minmu=None,
    parallel=False,
):
    """
    Differential expression analysis based on the Negative Binomial distribution.

    This function performs a default analysis through the steps:

    - estimation of size factors (:meth:`.DESeqDataSet.estimateSizeFactors`)
    - estimation of dispersion (:meth:`.DESeqDataSet.estimateDispersions`)
    - negative binomial GLM fitting and Wald statistics (:func:`nbinomWaldTest`
      or :func:`nbinomLRT`)

    For complete details on each step, see the documentation of the
    corresponding functions. After :func:`DESeq` returns a
    :class:`~DESeqDataSet.DESeqDataSet`, results tables (log2 fold change and
    p-values) can be generated using the
    :meth:`.DESeqDataSet.results` function.  Shrunken LFC can then
    be generated using :func:`lfcShrink`.

    Notes
    -----
    The differential expression analysis uses a generalized linear model of the
    form:

    .. math::

       K_{ij} &\sim \mathcal{NB}(\mu_{ij}, \\alpha_i)

       \mu_{ij} &\sim s_j \; q_{ij}

       \log q_{ij} &= x_{j.} \; \\beta_i

    where counts :math:`K_{ij}` for gene :math:`i` and sample :math:`j` are
    modeled using a Negative Binomial distribution with fitted mean
    :math:`\mu_{ij}` and a gene-specific dispersion parameter
    :math:`\\alpha_i`.  The fitted mean is composed of a sample-specific size
    factor :math:`s_j` and a parameter :math:`q_{ij}` proportional to the
    expected true concentration of framents for sample :math:`j`.  The
    coefficients :math:`\\beta_i` give the log2 fold changes for gene :math:`i`
    for each column of the model matrix :math:`X`.  The sample-specific size
    factors can be replaced by gene-specific normalization factors for each
    sample using :attr:`.DESeqDataSet.normalizationFactors`.

    For details on the fitting of the log2 fold changes and calculation of
    p-values, see :func:`nbinomWaldTest` if using :code:`test="Wald"`, or
    :func:`nbinomLRT` if using :code:`test="LRT"`.

    Experiments without replicates do not allow for estimation of the dispersion
    of counts around the expected value for each group, which is critical for
    differential expression analysis.

    The argument :code:`minReplicatesForReplace` is used to decide which
    samples are eligible for automatic replacement in the case of extreme
    Cook's distance.  By default, :func:`DESeq` will replace outliers if the
    Cook's distance is large for a sample which has 7 or more replicates
    (including itself). Outlier replacement is turned off entirely for
    :code:`fitType="glmGamPoi"`. This replacement is performed by
    :func:`replaceOutliers`. This default behavior helps to prevent filtering
    genes based on Cook's distance when there are many degrees of freedom. See
    :meth:`.DESeqDataSet.results` for more information about filtering using
    Cook's distance, and the "Dealing with outliers" section of the module
    documentation. Unlike the behavior of :func:`replaceOutliers`, here
    original counts are kept in the matrix returned by :code:`counts`, original
    Cook's distances are kept in :code:`obj.layers["cooks"]`, and the
    replacement counts used for fitting are kept in
    :code:`obj.layers["replaceCounts"]`.

    Note that is a log2 fold change prior is used (:code:`betaPrior=True`),
    then expanded model matrices will be used for fitting. These are described
    in :func:`nbinomWaldTest` and in the module documentation. The
    :code:`contrast` argument of :meth:`.DESeqDataSet.results` should be used
    for generating result tables.

    Arguments
    ---------
    obj : DESeqDataSet
        DESeqDataSet object
    test : str, optional
        either :code:`"Wald"` (default) or :code:`"LRT"`, to choose between
        Wald significance tests (:func:`nbinomWaldTest`) or the likelihood
        ratio test on the difference in deviance between a full and a reduced
        model formula (:func:`nbinomLRT`).
    fitType : str, optional
        either :code:`"parametric"` (default), :code:`"local"`, :code:`"mean"`
        or :code:`"glmGamPoi"` for the type of fitting of dispersions to the
        mean intensity. See :meth:`.DESeqDataSet.estimateDispersions` for
        details.
    sfType : str, optional
        either :code:`"ratio"` (default), :code:`"poscounts"` or
        :code:`"iterate"` for the type of size factor estimation. See
        :meth:`.DESeqDataSet.estimateSizeFactors` for details.
    betaPrior : bool, optional
        whether to put a zero-mean normal prior on the non-intercept
        coefficients. See :func:`nbinomWaldTest` for details on the calculation
        of the beta prior. Defaults to :code:`False`, and shrunken LFCs are
        obtained afterwards using :func:`lfcShrink`.
    full : formula or design matrix, optional
        for :code:`test="LRT"`, the full model formula, which is restricted to
        the formula in :code:`obj.design`. Alternatively, it can be a model
        matrix constructed by the user.
        Advanced use: specifying a model matric for full and
        :code:`test="Wald"` is possible if :code:`betaPrior=False`.
    reduced : formula or design matrix, optional
        for :code:`test="LRT"`, a reduced formula to compare against, i.e. the
        full formula with the terms of interest removed.  Alternatively, it can
        be a model matrix constructed by the user.
    quiet : bool
        whether to print messages at each step
    minReplicatesForReplace : int, optional
        the minimum number of replicates required in order to use
        :func:`replaceOutliers` on a sample.  If there are samples with that
        many replicates, the model will be refit after replacing
        outliers, flagged by Cook's distance.
        Set it to :code:`inf` in order to never replace outliers.
        It is set to :code:`inf` if :code:`fitType="glmGamPoi"`.
    modelMatrixType : str, optional
        either :code:`"standard"` or :code:`"expanded"`. Describes how the
        model matrix X of the GLM formula is formed.  If "standard", model
        matrix is built directly from the design formula.  If "expanded", model
        matrix includes an indicator variable for each level of factors in
        addition to an intercept.  :code:`betaPrior` must be :code:`True` in
        order for expanded model matrices to be fitted.
    useT : bool, optional
        passed to :func:`nbinomWaldTest`, where Wald statistics are assumed to
        follow a standard Normal
    minmu : float, optional
        lower bound on the estimated count for fitting gene-wise dispersion and
        for use with nbinomWaldTest and nbinomLRT.  If :code:`fitType ==
        "glmGamPoi"` then defaults to 1e-6 (as this :code:`fitType` is
        optimized for single cell data, for which a lower :code:`minmu` is
        recommended), otherwise defaults to 0.5
    parallel : bool
        unimplemented

    Returns
    -------
    DESeqDataSet
        the input :code:`obj`, updated with differential expression analysis
        data
    """

    # Default values
    if minmu is None:
        if fitType == "glmGamPoi":
            minmu = 1e-6
        else:
            minmu = 0.5

    # Check arguments
    if not isinstance(obj, DESeqDataSet):
        raise ValueError("obj is not of type DESeqDataSet")

    if test not in ["Wald", "LRT"]:
        raise ValueError("invalid value for parameter test. Must be either Wald or LRT")

    if fitType not in ["parametric", "local", "mean", "glmGamPoi"]:
        raise ValueError(f"invalid value for parameter fitType: {fitType}")

    dispersionEstimator = "glmGamPoi" if fitType == "glmGamPoi" else "DESeq2"

    if not isinstance(quiet, bool):
        raise ValueError(f"invalid value for parameter quiet: {quiet}")
    if not quiet:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)

    if not isinstance(parallel, bool):
        raise ValueError(f"invalid value for parameter parallel: {parallel}")

    # turn off outlier replacement for glmGamPoi
    if fitType == "glmGamPoi":
        minReplicatesForReplace = np.inf
        if parallel:
            logging.warnings.warn(
                "parallelization of DESeq() is not implemented for fitType='glmGamPoi'"
            )

    if sfType not in ["ratio", "poscounts", "iterate"]:
        raise ValueError(f"invalid value for parameter 'sfType': {sfType}")

    # more argument checking
    # TODO check that minReplicatesForReplace is numeric

    modelAsFormula = not isinstance(full, (np.ndarray, pd.DataFrame))

    if not isinstance(betaPrior, bool):
        raise ValueError(f"invalid value for parameter betaPrior: {betaPrior}")

    if full is None:
        full = obj.design
    else:
        full = patsy.dmatrix(full, data=obj.obs, NA_action="raise")

    if test == "LRT":
        if reduced is None:
            raise ValueError("likelihood ratio test requires a 'reduced' design")
        if betaPrior:
            raise ValueError(
                "test='LRT' does not support use of LFC shrinkage, use betaPrior=False"
            )
        if modelMatrixType is not None and modelMatrixType == "expanded":
            raise ValueError("test='LRT' does not support use of expanded model matrix")

        if modelAsFormula == isinstance(reduced, (np.ndarray, pd.DataFrame)):
            raise ValueError(
                "if one of 'full' or 'reduced' is a matrix, the other must also be a matrix"
            )

        if modelAsFormula:
            reduced = patsy.dmatrix(reduced, data=obj.obs, NA_action="raise")
            checkLRT(full, reduced)
        else:
            checkFullRank(full)
            checkFullRank(reduced)
            if full.shape[1] <= reduced.shape[1]:
                raise ValueError(
                    "the number of columns of 'full' should be larger than the number of columns of 'reduced'"
                )

    if test == "Wald" and reduced is not None:
        raise ValueError("'reduced' ignored when test='Wald'")
    if dispersionEstimator == "glmGamPoi" and test == "Wald":
        logging.warnings.warn(
            "glmGamPoi dispersion estimator should be used in combination with a LRT and not a Wald test"
        )

    if modelAsFormula:
        # run some tests common to DESeq, nbinomWaldTest, nbinomLRT
        obj.designAndArgChecker(betaPrior)

        # warn if the design is just an intercept
        if [len(t.factors) for t in obj.design.design_info.terms] == [0]:
            logging.warnings.warn(
                "the design is ~1 (just an intercept). Is this intended?"
            )

        if full.design_info.describe() != obj.design.design_info.describe():
            raise ValueError("'full' specified as formula should match obj.design")

        modelMatrix = None

    else:
        # model not as formula, so DESeq() is using supplied model matrix
        logging.info("using supplied model matrix")
        if betaPrior:
            raise ValueError(
                "'betaPrior'=True is not supported for user-provided model matrices"
            )
        checkFullRank(full)
        # this will be used for dispersion estimation and testing
        modelMatrix = full

    obj.betaPrior = betaPrior

    if obj.normalizationFactors is not None:
        logging.info("using pre-existing normalization factors")
    elif obj.sizeFactors is not None:
        logging.info("using pre-existing size factors")
    else:
        logging.info("estimating size factors")
        obj = obj.estimateSizeFactors(type_=sfType, quiet=quiet)

    if not parallel:
        logging.info("estimating dispersions")

        obj = obj.estimateDispersions(
            fitType=fitType, quiet=quiet, modelMatrix=modelMatrix, minmu=minmu
        )

        logging.info("fitting model and testing")

        if test == "Wald":
            obj = nbinomWaldTest(
                obj,
                betaPrior=betaPrior,
                quiet=quiet,
                modelMatrix=modelMatrix,
                modelMatrixType=modelMatrixType,
                useT=useT,
                minmu=minmu,
            )
        elif test == "LRT":
            obj = nbinomLRT(
                obj,
                full=full,
                reduced=reduced,
                quiet=quiet,
                minmu=minmu,
                type_=dispersionEstimator,
            )

    else:  # if parallel
        if modelMatrixType is not None:
            if betaPrior and modelMatrixType != "expanded":
                raise ValueError(
                    "parallelization not implemented for non-expanded matrix with beta priors"
                )

        obj = DESeqParallel(
            obj,
            test=test,
            fitType=fitType,
            betaPrior=betaPrior,
            full=full,
            reduced=reduced,
            quiet=quiet,
            modelMatrix=modelMatrix,
            useT=useT,
            minmu=minmu,
            BPPARAM=BPPARAM,
        )

    # if there are sufficient replicates, then pass through to refitting function
    sufficientReps = nOrMoreInCell(obj.modelMatrix, minReplicatesForReplace).any()
    if sufficientReps:
        obj = refitWithoutOutliers(
            obj,
            test=test,
            betaPrior=betaPrior,
            full=full,
            reduced=reduced,
            quiet=quiet,
            minReplicatesForReplace=minReplicatesForReplace,
            modelMatrix=modelMatrix,
            modelMatrixType=modelMatrixType,
        )

    # TODO R DESeq2 stores the package version in obj

    return obj



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/mglmOneWay.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np
from ..utils import Factor, asfactor
from .makeCompressedMatrix import (
    _compressOffsets,
    _compressDispersions,
    _compressWeights,
)
from .mglmOneGroup import mglmOneGroup
from .edgepy_cpp import cxx_get_one_way_fitted
from scipy.linalg import solve


def designAsFactor(design):
    """
    Convert a general design matrix into a oneway layout if that is possible.

    This function determines how many distinct row values the design matrix is
    capable of computing and returns a factor with a level for each possible
    distinct value.

    Arguments
    ---------
    design : matrix
        the design matrix. Assumed to be full column rank.

    Returns
    -------
    Factor
        factor of length equal to the number of rows in :code:`design`
    """
    design = np.asarray(design, order="F")
    z = (np.e + np.pi) / 5
    col = np.full(design.shape, np.arange(design.shape[1]), order="F")
    means = (design * z**col).mean(axis=1)
    uniq = np.unique(means)
    fact = np.zeros(means.shape, dtype="int64", order="F")
    for i in range(len(uniq)):
        fact[means == uniq[i]] = i + 1
    return Factor(fact)


def mglmOneWay(
    y,
    design=None,
    group=None,
    dispersion=0,
    offset=0,
    weights=None,
    coef_start=None,
    maxit=50,
    tol=1e-10,
):
    """
    Fit multiple negative binomial GLMs with log-link by Fisher scoring with a
    single explanatory factor in the model.

    This is a low-level work-horse used by higher-level functions, especially
    :func:`glmFit`. Most users will not need to call this function directly.

    This function fits a negative binomial GLM to each row of :code:`y`. The
    row-wise GLMs all have the same design matrix but possibly different
    dispersions, offsets and weights. It is low-level in that it
    operates on atomic objects (matrices and vectors).

    This function fits a oneway layout to each response vector. It treats the
    libraries as belonging to a number of groups and calls :func:`mglmOneGroup`
    for each group. It treats the dispersion parameter of the negative binomial
    distribution as a known input.

    Arguments
    ---------
    y : array_like
        matrix of negative binomial counts. Rows for genes and columns for
        libraries.
    design : array_like, optional
        design matrix of the GLM. Assumed to be full column rank. Defaults to
        :code:`~ 0 + group` if :code:`group` is specified, otherwise to
        :code:`~ 1`.
    group : :obj:`Factor`
        group memberships for oneway layout. If both :code:`design` and
        :code:`group` are specified, then they must agree in terms of
        :func:`designAsFactor`. If :code:`design = None`, then a group-means
        design matrix is implied.
    dispersion : float or array_like
        scalar or vector giving the dispersion parameter for each GLM. Can be a
        scalar giving one value for all genes, or a vector of length equal to
        the number of genes giving genewise dispersions.
    offset : array_like
        vector or matrix giving the offset that is to be included in the log
        linear model predictor. Can be a scalar, a vector of length equal to the
        number of libraries, or a matrix of the same shape as :code:`y`.
    weights : matrix, optional
        vector or matrix of non-negative quantitative weights. Can be a vector
        of length equal to the number of libraries, or a matrix of the same
        shape as :code:`y`.
    coef_start : array_like, optional
        matrix of starting values for the linear model coefficient. Number of
        rows should agree with :code:`y` and number of columns should agree with
        :code:`design`. This argument does not usually need to be set as the
        automatic starting values perform well.
    maxit : int
        the maximum number of iterations for the Fisher scoring algorithm. The
        iteration will be stopped when this limit is reached even if the
        convergence criterion has not been satisfied.
    tol : float
        the convergence tolerance.

    Returns
    -------
    tuple
        tuple with the following components:

        - matrix of estimated coefficients for the linear models. Rows
          corrspond to row of :code:`y` and columns to columns of :code:`design`
        - matrix of fitted values. Same shape as :code:`y`.
    """
    y = np.asarray(y, order="F")
    (ngenes, nlibs) = y.shape

    offset = _compressOffsets(y, offset=offset)
    dispersion = _compressDispersions(y, dispersion)
    weights = _compressWeights(y, weights)

    # If necessary, the group factor is computed from the design matrix.
    # However, if group is supplied, we can avoid creating a design matrix altogether.
    if group is None:
        if design is None:
            group = Factor(np.ones((nlibs,)))
        else:
            design = np.asarray(design, order="F")
            group = designAsFactor(design)
    else:
        group = asfactor(group)

    # Convert factor to integer levels for efficiency
    levg = group.categories
    ngroups = len(levg)
    i = group.__array__()

    if design is not None:
        if design.shape[1] != ngroups:
            raise ValueError("design matrix is not equivalent to a oneway layout")

    # Reduce to representative design matrix, based on the column in which each group appears first
    firstjofgroup = [(i == x).nonzero()[0][0] for x in levg]
    if design is not None:
        designunique = design[firstjofgroup, :]
    else:
        designunique = None

    # Is it just a group indicator matrix?
    if (
        np.sum(designunique == 1) == ngroups
        and np.sum(designunique == 0) == (ngroups - 1) * ngroups
    ):
        design = None

    # If necessary, convert starting values to group fitted values
    if design is not None and coef_start is not None:
        coef_start = coef_start @ designunique.T

    # Cycle through groups
    beta = np.zeros((ngenes, ngroups), order="F", dtype="double")
    for g in range(ngroups):
        j = np.nonzero(i == levg[g])[0]
        beta[:, g] = mglmOneGroup(
            y[:, j],
            dispersion=dispersion[:, j],
            offset=offset[:, j],
            weights=weights[:, j] if weights is not None else None,
            coef_start=coef_start[:, g] if coef_start is not None else None,
            maxit=maxit,
            tol=tol,
        )

    # Reset -inf values to finite values to simplify calculations downstream
    beta = np.where(beta > -1e8, beta, -1e8)

    # Fitted values from group-wise beta's
    mu = cxx_get_one_way_fitted(beta, offset, i - 1)

    # If necessary, reformat the beta's to reflect the original design.
    if design is not None:
        beta = solve(designunique, beta.T).T

    return (beta, mu)



import unittest
import numpy as np

from inmoose.edgepy import DGEList, validDGEList


class test_utils(unittest.TestCase):
    def test_isAllZero(self):
        from inmoose.edgepy.utils import _isAllZero

        self.assertEqual(_isAllZero(np.array([])), False)
        self.assertEqual(_isAllZero(np.array([0, 0, 0])), True)
        self.assertEqual(_isAllZero(np.array([2, 1, 0])), False)
        arr = np.array([np.NaN, 0, -1, np.PINF, 1, np.NINF])
        with self.assertRaisesRegex(
            ValueError, expected_regex="NaN counts are not allowed"
        ):
            _isAllZero(np.array([1, 2, 3, np.NaN, 4, 5]))
        with self.assertRaisesRegex(
            ValueError, expected_regex="infinite counts are not allowed"
        ):
            _isAllZero(np.array([1, 2, 3, np.PINF, 4, 5]))
        with self.assertRaisesRegex(
            ValueError, expected_regex="negative counts are not allowed"
        ):
            _isAllZero(np.array([1, 2, 3, np.NINF, 4, 5]))
        with self.assertRaisesRegex(
            ValueError, expected_regex="negative counts are not allowed"
        ):
            _isAllZero(np.array([1, 2, 3, -42, 4, 5]))


class test_DGEList(unittest.TestCase):
    def test_constructor(self):
        with self.assertRaisesRegex(
            ValueError, expected_regex="'counts' is not a matrix"
        ):
            d = DGEList(None)
        with self.assertRaisesRegex(
            ValueError, expected_regex="non-numeric values found in 'counts'"
        ):
            d = DGEList([["foobar"]])
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="length of 'lib_size' must be equal to the number of columns in 'counts'",
        ):
            d = DGEList([[42]], lib_size=[1, 2])
        with self.assertRaisesRegex(
            ValueError, expected_regex="negative library size not permitted"
        ):
            d = DGEList([[42]], lib_size=[-1])
        with self.assertRaisesRegex(
            ValueError, expected_regex="library size set to zero but counts are nonzero"
        ):
            d = DGEList([[42]], lib_size=[0])
        with self.assertWarnsRegex(
            UserWarning, expected_regex="library size of zero detected"
        ):
            d = DGEList([[0]], lib_size=[0])

        d = DGEList([[42]])
        self.assertIsNotNone(d.counts)
        self.assertIsNotNone(d.samples)
        self.assertIsNone(d.common_dispersion)
        self.assertIsNone(d.trended_dispersion)
        self.assertIsNone(d.tagwise_dispersion)
        self.assertIsNone(d.design)
        self.assertIsNone(d.weights)
        self.assertIsNone(d.offset)
        self.assertIsNone(d.genes)
        self.assertIsNone(d.prior_df)
        self.assertIsNone(d.AveLogCPM)

    def test_getOffset(self):
        d = DGEList([[42]])
        self.assertTrue(np.allclose(d.getOffset(), 3.73767, atol=1e-6, rtol=0))
        d.offset = 0
        self.assertEqual(d.getOffset(), 0)

    def test_getDispersion(self):
        d = DGEList([[42]])
        self.assertIsNone(d.getDispersion())
        d.common_dispersion = "common"
        self.assertEqual(d.getDispersion(), "common")
        d.trended_dispersion = "trended"
        self.assertEqual(d.getDispersion(), "trended")
        d.tagwise_dispersion = "tagwise"
        self.assertEqual(d.getDispersion(), "tagwise")

    def test_validDGEList(self):
        d = DGEList([[42]])
        d.counts = None
        with self.assertRaisesRegex(RuntimeError, expected_regex="No count matrix"):
            validDGEList(d)
        d = DGEList([[42]])
        d = validDGEList(d)
        self.assertFalse((d.samples.group.values == None).any())
        self.assertFalse((d.samples.lib_size.values == None).any())
        self.assertFalse((d.samples.norm_factors.values == None).any())



import unittest
import numpy as np
from pandas.api.types import is_categorical_dtype

from inmoose.deseq2 import DESeqDataSet


class Test(unittest.TestCase):
    def test_counts(self):
        """test that normalized counts are properly computed"""
        dds = DESeqDataSet(
            np.arange(24).reshape(4, 6), clinicalData=["A", "A", "B", "B"], design="~1"
        )
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="first calculate size factors, add normalizationFactors, or set normalized=False",
        ):
            dds.counts(normalized=True)
        dds = dds.estimateSizeFactors()
        ref = np.array(
            [
                [0.000000, 3.201086, 6.402172, 9.603258, 12.80434, 16.00543],
                [6.402172, 7.469200, 8.536229, 9.603258, 10.67029, 11.73731],
                [7.682606, 8.322823, 8.963040, 9.603258, 10.24347, 10.88369],
                [8.231364, 8.688662, 9.145960, 9.603258, 10.06056, 10.51785],
            ]
        )
        res = dds.counts(normalized=True)
        self.assertTrue(np.allclose(res, ref))

    def test_design(self):
        """test that categorical variable in the design are properly accounted for"""
        dds = DESeqDataSet(np.arange(24).reshape(4, 6))
        dds.obs["x"] = ["A", "A", "B", "B"]
        dds.obs["y"] = [1, 2, 1, 2]
        dds.design = "x + y"
        self.assertTrue("C(x)" in dds.obs)
        self.assertTrue("C(y)" not in dds.obs)
        self.assertFalse(is_categorical_dtype(dds.obs["x"].dtype))
        self.assertTrue(is_categorical_dtype(dds.obs["C(x)"].dtype))
        self.assertFalse(is_categorical_dtype(dds.obs["y"].dtype))

        dds = DESeqDataSet(np.arange(24).reshape(4, 6))
        dds.obs["x"] = ["A", "A", "B", "B"]
        dds.obs["y"] = [1, 2, 1, 2]
        dds.design = "C(x) + C(y)"
        self.assertTrue("C(x)" in dds.obs)
        self.assertTrue("C(y)" in dds.obs)
        self.assertFalse(is_categorical_dtype(dds.obs["x"].dtype))
        self.assertTrue(is_categorical_dtype(dds.obs["C(x)"].dtype))
        self.assertFalse(is_categorical_dtype(dds.obs["y"].dtype))
        self.assertTrue(is_categorical_dtype(dds.obs["C(y)"].dtype))



import unittest
import numpy as np

from inmoose.deseq2 import DESeqDataSet


class Test(unittest.TestCase):
    @unittest.skip("fpkm not implemented")
    def test_fpkm(self):
        """test that fpkm works as expected"""
        df = {"x": [1, 2]}
        dds = DESeqDataSet(np.array([[1, 2, 3, 4], [2, 4, 6, 8]]), df, "~1")
        self.assertEqual(dds.fpkm()[0, 0], 1e7)
        self.assertEqual(dds.fpm()[0, 0], 1e5)
        self.assertEqual(dds.fpm(robuts=False)[0, 0], 1e5)



import importlib.metadata

__version__ = importlib.metadata.version(__package__)

# force RTLD_GLOBAL when loading common_cpp
from sys import getdlopenflags, setdlopenflags
from os import RTLD_GLOBAL

default_dlopen_flags = getdlopenflags()
setdlopenflags(default_dlopen_flags | RTLD_GLOBAL)
from . import common_cpp

setdlopenflags(default_dlopen_flags)
del default_dlopen_flags

from . import edgepy
from . import pycombat
from . import utils
from . import consensus_clustering
from . import deseq2

import logging

logging.basicConfig(level=logging.INFO, format="[{levelname}] {message}", style="{")
logging.captureWarnings(True)



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2023 A. Behdenna, A. Nordor, J. Haziza, A. Gema, M. Colange, L. Meunier

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import logging
import numpy as np
import pandas as pd
from patsy import dmatrix, DesignMatrix

from ..utils import asfactor


class VirtualCohortInput:
    """
    Attributes
    ----------
    counts : matrix
        raw count matrix from genomic studies (dimensions gene x sample)
    batch : :class:`~inmoose.utils.factor.Factor`
        batch identifiers
    nan_batch : bool
        whether :code:`batch` contains :code:`NA` values
    nan_genes : int
        number of genes with a :code:`NA` value in the :code:`counts` matrix
    nan_samples : int
        number of samples with a :code:`NA` value in the :code:`counts` matrix
    batch_composition : dict[batch_id, list[int]]
        for each batch, the indices of the samples of this batch
    n_batch : int
        number of batches
    ref_batch_idx : int or None
        index (in :code:`batch.categories`) of the reference batch
    nan_cov : int
        number of covariates containing a :code:`NA` value
    nan_cov_samples : dict[batch_id, int]
        number of samples with a :code:`NA` covariate per batch
    confounded_cov : Option[list]
        list of indices of confounded variables
        :code:`None` if some covariates contain :code:`NA` values, because
        confounded covariates cannot be determined
    batch_mod : :class:`patsy.DesignMatrix`
        the batch-only design matrix (i.e. without covariates)
        :code:`None` if some batch is :code:`NA`
    covar_mod : :class:`patsy.DesignMatrix` or :class:`pd.DataFrame`
        the covariate-only design matrix (i.e. without batches)
        as a :class:`pd.DataFrame` only if some covariates contain :code:`NA` values
    design : :class:`patsy.DesignMatrix`
        the design matrix (combining batches and covariates)
        :code:`None` if some covariates contain :code:`NA` values or if some
        batch is :code:`NA`
    """

    def __init__(self, counts, batch, covar_mod=None, ref_batch=None):
        """
        Arguments
        ---------
        counts : matrix
            raw count matrix from genomic studies (dimensions gene x sample)
        batch : array or list or :class:`~inmoose.utils.factor.Factor`
            batch identifiers
        covar_mod : list or matrix, optional
            model matrix (dataframe, list or numpy array) for one or multiple
            covariates to include in linear model (signal from these variables
            are kept in data after adjustment). Covariates have to be
            categorial, they can not be continious values (default: :code:`None`).
        ref_batch : any
            batch id of the batch to use as reference. Must be one of the element of
            :code:`batch` (default: :code:`None`).
        """
        if isinstance(counts, pd.DataFrame):
            list_samples = counts.columns
            list_genes = counts.index
            counts = counts.values
        elif isinstance(counts, np.ndarray):
            list_samples = None
            list_genes = None
        else:
            raise ValueError("counts must be a pandas DataFrame or a numpy nd array")

        self.list_samples = list_samples
        self.list_genes = list_genes
        self.counts = counts
        self.batch = asfactor(batch)
        self.nan_batch = self.batch.isna().any()

        nancounts = np.isnan(counts)
        self.nan_genes = nancounts.any(axis=1).sum()
        self.nan_samples = nancounts.any(axis=0).sum()

        # number of batches
        self.n_batch = self.batch.nlevels()
        # number of samples in each batch
        self.batch_composition = {
            b: np.where(self.batch == b)[0] for b in self.batch.categories
        }

        if self.nan_batch:
            self.batch_mod = None
        else:
            # batch design matrix
            self.batch_mod = dmatrix("~0 + C(self.batch)")

        # reference batch
        if ref_batch is not None:
            if ref_batch not in self.batch.categories:
                raise ValueError("Reference batch must identify one of the batches")
            # ref_batch_idx is the index of the reference batch in batch.categories
            self.ref_batch_idx = np.where(self.batch.categories == ref_batch)[0][0]
            # update batchmod with reference
            self.batch_mod[:, self.ref_batch_idx] = 1
        else:
            self.ref_batch_idx = None

        if covar_mod is not None and not isinstance(covar_mod, DesignMatrix):
            # convert to dataframe
            if isinstance(covar_mod, (list, np.ndarray)):
                covar_mod = pd.DataFrame(covar_mod)
            elif type(covar_mod) != pd.DataFrame:
                raise ValueError(
                    f"The covar_mod parameter type {type(covar_mod)} is not accepted, it must be list, numpy.array or pandas.dataframe."
                )

            if covar_mod.shape[0] != len(self.batch):
                if covar_mod.shape[1] == len(self.batch):
                    logging.warn(
                        "The covariate matrix seems to be transposed. The computation will proceed with the input covariate matrix transposed, but you should double-check the covariate matrix."
                    )
                    covar_mod = covar_mod.T
                else:
                    raise ValueError(
                        "The covar_mod matrix parameter doesn't match the number of sample"
                    )

            if covar_mod.columns.values.dtype == "int64":
                covar_mod.columns = [
                    "cov_" + str(col_nb) for col_nb in covar_mod.columns
                ]

            nan_covar_mod = covar_mod.isna()
            self.nan_cov = nan_covar_mod.any(axis=0).sum()
            self.nan_cov_samples = {
                b: nan_covar_mod[self.batch == b].any(axis=0).sum()
                for b in self.batch.categories
            }
        else:
            self.nan_cov = 0
            self.nan_cov_samples = {b: 0 for b in self.batch.categories}

        if self.nan_cov == 0:
            if covar_mod is not None:
                if not isinstance(covar_mod, DesignMatrix):
                    covar_mod = dmatrix(
                        "+".join([f"{cv}" for cv in covar_mod.columns]),
                        data=covar_mod,
                    )
            else:
                covar_mod = dmatrix("~1", pd.DataFrame(counts.T))

            self.confounded_cov = []
            for c in range(covar_mod.shape[1]):
                # skip the intercept
                if (covar_mod[:, c] == 1).all():
                    continue
                if (
                    np.linalg.matrix_rank(
                        np.hstack([self.batch_mod, covar_mod[:, c][:, None]])
                    )
                    <= self.batch_mod.shape[1]
                ):
                    self.confounded_cov.append(c)

        else:
            self.confounded_cov = None

        self.covar_mod = covar_mod
        if self.nan_cov == 0 and not self.nan_batch:
            self.design = dmatrix("~ 0 + self.batch_mod + covar_mod")
            # Check for intercept in covariates, and drop if present
            check = (self.design == 1).all(axis=0)
            if self.ref_batch_idx is not None:
                # the reference batch is not considered as a covariate
                check[self.ref_batch_idx] = False
            self.design = self.design[:, np.logical_not(check)]
        else:
            self.design = None

    def run_check(self, na_cov_action="raise"):
        """
        Helper function to control the quality of the input virtual cohort.

        Arguments
        ---------
        na_cov_action : {"raise", "remove", "fill"}, optional
            the action to take to handle :code:`NA` values in covariates:
            - :code:`"raise"` do nothing and raise a :code:`ValueError`
            - :code:`"remove"` remove the samples impacted by the :code:`NA` covariates
            - :code:`"fill"` replace the :code:`NA` by a new covariate category per batch.
              not applicable to non-categorical covariates (an error will be raised)
            defaults to "raise"

        Returns
        -------
        bool
            whether the input was modified
        VirtualCohortInput
            the resulting input after modification (if any)

        Examples
        --------
        Typical usage is as follows:
        >>> while True:
        ...     (modified, vc_input) = VirtualCohortInput(counts, batch, covar)
        ...     if not modified:
        ...         break
        """
        if self.nan_batch:
            raise ValueError(
                "NaN batches detected! Please review your batches before proceeding with batch effect correction."
            )

        single_sample_batches = [
            b for b, v in self.batch_composition.items() if len(v) == 1
        ]
        if len(single_sample_batches) > 0:
            raise ValueError(
                f"Batches {', '.join([str(b) for b in single_sample_batches])} contain a single sample, which is not supported for batch effect correction. Please review your inputs."
            )

        if self.nan_genes > 0 or self.nan_samples > 0:
            raise ValueError(
                f"The count matrix contains NaN values impacting {self.nan_genes} genes and {self.nan_samples} samples. Please review your inputs."
            )

        if self.nan_cov > 0:
            counts, batch, covar = self.fix_na_cov(na_cov_action)
            if self.ref_batch_idx is not None:
                ref_batch = self.batch[self.ref_batch_idx]
            else:
                ref_batch = None
            return (
                True,
                VirtualCohortInput(counts, batch, covar_mod=covar, ref_batch=ref_batch),
            )

        if self.confounded_cov is not None and len(self.confounded_cov) > 0:
            raise ValueError(
                f"Covariates {', '.join([str(self.covar_mod.design_info.column_names[c]) for c in self.confounded_cov])} are confounded with the batches. Please review your covariates before proceeding with batch effect correction."
            )
        return (False, self)

    def fix_na_cov(self, na_cov_action="raise"):
        """
        A helper function to handle :code:`NA` in covariates

        Arguments
        ---------
        counts : matrix
            raw count matrix from genomic studies (dimensions gene x sample)
        batch : :class:`~inmoose.utils.factor.Factor`
            batch identifiers
        covar_mod : :class:`patsy.DesignMatrix`
            covariate matrix
        na_cov_action : {"raise", "remove", "fill"}, optional
            the action to take:
            - :code:`"raise"` do nothing and raise a :code:`ValueError`
            - :code:`"remove"` remove the samples impacted by the :code:`NA` covariates
            - :code:`"fill"` replace the :code:`NA` by a new covariate category per batch.
              not applicable to non-categorical covariates (an error will be raised)
            defaults to "raise"

        Returns
        -------
        matrix
            resulting count matrix
        :class:`~inmoose.utils.factor.Factor`
            resulting batch identifiers
        :class:`patsy.DesignMatrix`
            resulting covariate Matrix
        """
        nan_covar_mod = self.covar_mod.isna()
        if na_cov_action == "raise":
            name_nan_covar_mod = nan_covar_mod.loc[
                :, nan_covar_mod.any() == True
            ].columns
            raise ValueError(
                f"{nan_covar_mod.sum().sum()} values are missing in covariates {', '.join(name_nan_covar_mod)}. Correct your covariates or use the cov_missing_value parameters"
            )
        elif na_cov_action == "remove":
            logging.warnings.warn(
                f"{(nan_covar_mod.sum(axis=1)>0).sum()} samples with missing covariates in covar_mod. They are removed from the data. You may want to double check your covariates."
            )
            keep = nan_covar_mod.sum(axis=1) == 0
            return self.counts[:, keep], self.batch[keep], self.covar_mod[keep]
        elif na_cov_action == "fill":
            logging.warnings.warn(
                f"{nan_covar_mod.sum().sum()} missing covariates in covar_mod. Creating a distinct covariate per batch for the missing values. You may want to double check your covariates."
            )

            # check for numeric type (excluding nan) to identify potential continuous_variable
            continuous_variable = []
            potential_continuous_variable = []
            for col in self.covar_mod.columns:
                # Return True if numeric type with decimal
                # Return False if numeric type without decimal
                col_data_type = [
                    (ele % 1) > 0
                    for ele in set(self.covar_mod[col].dropna())
                    if isinstance(ele, (float, int, complex))
                ]
                if True in col_data_type:
                    continuous_variable.append(col)
                elif False in col_data_type:
                    potential_continuous_variable.append(col)

            nan_cols = self.covar_mod.columns[self.covar_mod.isna().any(axis=0)]
            continuous_nan_cols = np.intersect1d(nan_cols, continuous_variable)
            maybe_continuous_nan_cols = np.intersect1d(
                nan_cols, potential_continuous_variable
            )
            if len(continuous_nan_cols) > 0:
                raise ValueError(
                    f"Cannot create new categories for numerical covariates {', '.join(continuous_nan_cols)}. Please fix the NA in those covariates manually."
                )
            if len(maybe_continuous_nan_cols) > 0:
                logging.warnings.warn(
                    f"Creating new categories for integer covariates {', '.join(maybe_continuous_nan_cols)}. These are treated as categorical covariates, but you may want to double-check those."
                )

            # handle missing covariates, by creating a distinct covariate per batch
            # where a missing covariate appears
            covar_mod = self.covar_mod
            for col in self.covar_mod.columns:
                nan_cov_col = self.covar_mod[col].isna()
                nan_batch_group = [
                    f"nan_batch_{self.batch[i]}"
                    for i in range(len(self.covar_mod[col]))
                    if nan_cov_col[i]
                ]
                for i, j in enumerate(np.where(nan_cov_col)[0]):
                    covar_mod.loc[j, col] = nan_batch_group[i]
            return self.counts, self.batch, covar_mod
        else:
            raise ValueError(
                f"unknown value {na_cov_action} for argument 'na_cov_action': must be one of 'raise', 'remove' or 'fill'"
            )


def make_design_matrix(
    counts, batch, covar_mod=None, ref_batch=None, na_cov_action="raise"
):
    """Make design matrix for batch effect correction. Handles covariates as
    well as reference batch.

    Arguments
    ---------
    counts : matrix
        raw count matrix from genomic studies (dimensions gene x sample)
    batch : array or list or :obj:`inmoose.utils.factor.Factor`
        batch indices
    covar_mod : list or matrix, optional
        model matrix (dataframe, list or numpy array) for one or multiple
        covariates to include in linear model (signal from these variables are
        kept in data after adjustment). Covariates have to be categorial, they
        can not be continious values (default: `None`).
    ref_batch : any
        batch id of the batch to use as reference. Must be one of the element of
        `batch` (default: `None`).
    na_cov_action : {"raise", "remove", "fill"}, optional
        the action to take to handle :code:`NA` values in covariates:
        - :code:`"raise"` do nothing and raise a :code:`ValueError`
        - :code:`"remove"` remove the samples impacted by the :code:`NA` covariates
        - :code:`"fill"` replace the :code:`NA` by a new covariate category per batch.
          not applicable to non-categorical covariates (an error will be raised)
        defaults to "raise"

    Returns
    -------
    VirtualCohortInput
        a class containing all the information required for batch effect correction
    """

    vc_input = VirtualCohortInput(counts, batch, covar_mod, ref_batch)
    while True:
        (modified, vc_input) = vc_input.run_check(na_cov_action)
        if not modified:
            break

    logging.info(f"Found {vc_input.n_batch} batches")

    if ref_batch is not None:
        logging.info(f"Using batch {ref_batch} as reference batch")

    logging.info(
        f"Adjusting for {vc_input.design.shape[1] - vc_input.batch_mod.shape[1]} covariate(s) or covariate level(s)"
    )

    return vc_input



# distutils: language = c++
#-----------------------------------------------------------------------------
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

from libcpp cimport bool, pair, vector
from numpy cimport ndarray

# cf. file src/utils.h from edgeR source repo

cdef public ndarray[double, ndim=1] vector2ndarray "vector2ndarray"(const vector.vector[double]& data)

cdef extern from "objects.h":
    cpdef bool is_integer_array(ndarray arr) except +

cdef extern from "add_prior_count.h":
    # y dtype is either int or double
    cpdef object cxx_add_prior_count "add_prior_count"(
            ndarray y,
            ndarray[double] offset,
            ndarray[double] prior,
            ) except +

cdef extern from "compute_nbdev.cpp":
    # y dtype is either int or double
    cpdef vector.vector[double] cxx_compute_nbdev_sum "compute_nbdev_sum"(
            ndarray y,
            ndarray[double] mu,
            ndarray[double] phi,
            ndarray[double] weights) except +
    # y dtype is either int or double
    cpdef ndarray[double] cxx_compute_nbdev_nosum "compute_nbdev_nosum"(
            ndarray y,
            ndarray[double] mu,
            ndarray[double] phi,
            ndarray[double] weights) except +

cdef extern from "maximize_interpolant.cpp":
    cpdef vector.vector[double] cxx_maximize_interpolant "maximize_interpolant"(vector.vector[double] spts, ndarray[double] likelihoods) except +

cdef extern from "get_one_way_fitted.cpp":
    cpdef ndarray[double] cxx_get_one_way_fitted "get_one_way_fitted"(
            ndarray[double] beta,
            ndarray[double] offset,
            vector.vector[int] groups) except +

cdef extern from "fit_one_group.cpp":
    # y dtype is either int or double
    cpdef pair.pair[vector.vector[double], vector.vector[char]] cxx_fit_one_group "fit_one_group"(
            ndarray y,
            ndarray[double] offsets,
            ndarray[double] disp,
            ndarray[double] weights,
            long max_iterations,
            double tolerance,
            vector.vector[double] beta) except +

cdef extern from "ave_log_cpm.cpp":
    # y dtype is either int or double
    cpdef vector.vector[double] cxx_ave_log_cpm "ave_log_cpm"(
            ndarray y,
            ndarray[double] offset,
            ndarray[double] prior,
            ndarray[double] disp,
            ndarray[double] weights,
            long max_iterations,
            double tolerance) except +

# Helper function to return from C++ fit_levenberg function
cdef public object make_levenberg_result "make_levenberg_result"(
        ndarray[double, ndim=2] coefficients,
        ndarray[double, ndim=2] fitted_values,
        vector.vector[double] deviance,
        vector.vector[long] iter,
        vector.vector[char] failed)

cdef extern from "fit_levenberg.h":
    # y dtype is either int or double
    cpdef object cxx_fit_levenberg "fit_levenberg"(
            ndarray y,
            ndarray[double] offset,
            ndarray[double] disp,
            ndarray[double] weights,
            ndarray[double] design,
            ndarray[double] beta,
            double tol,
            long maxit) except +

cdef extern from "compute_apl.cpp":
    # y dtype is either int or double
    cpdef vector.vector[double] cxx_compute_apl "compute_apl"(
            ndarray y,
            ndarray[double] means,
            ndarray[double] disps,
            ndarray[double] weights,
            bool adjust,
            ndarray[double] design) except +

cdef extern from "initialize_levenberg.cpp":
    # y dtype is either int or double
    cpdef ndarray cxx_get_levenberg_start "get_levenberg_start"(
            ndarray y,
            ndarray[double] offset,
            ndarray[double] disp,
            ndarray[double] weights,
            ndarray[double] design,
            bool use_null) except +

cdef extern from "initialize_levenberg.h":
    cdef cppclass QRdecomposition:
        int NR
        int NC
        double* X
        vector.vector[double] Xcopy
        vector.vector[double] tau
        vector.vector[double] effects
        vector.vector[double] weights
        vector.vector[double] work_geqrf
        vector.vector[double] work_ormqr
        int lwork_geqrf
        int lwork_ormqr
        int info

# Indicate the C name to workaround a Cython bug
# see https://github.com/cython/cython/issues/2940
cdef public void build_QRdecomposition "build_QRdecomposition"(QRdecomposition* self)
cdef public void decompose_QRdecomposition "decompose_QRdecomposition"(QRdecomposition* self)
cdef public void solve_QRdecomposition "solve_QRdecomposition"(QRdecomposition* self)

# Re-expose in C BLAS/LAPACK functions exposed in Cython by scipy
cdef public void f77_dgemv "f77_dgemv"(const char *trans, const int *m, const int *n, const double *alpha, const double *a, const int *lda, const double *x, const int *incx, const double *beta, double *y, const int *incy) nogil
cdef public void f77_dpotrf "f77_dpotrf"(const char *uplo, const int *n, double *a, const int *lda, int *info) nogil
cdef public void f77_dpotrs "f77_dpotrs"(const char *uplo, const int *n, const int *nrhs, const double *a, const int *lda, double *b, const int *ldb, int *info) nogil
cdef public void f77_dsytrf "f77_dsytrf"(const char *uplo, const int *n, double *a, const int *lda, int *ipiv, double *work, const int *lwork, int *info) nogil




# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/core.R' of the Bioconductor DESeq2 package
# (version 3.16).


import numpy as np
import pandas as pd
import scipy.stats

from .Hmisc import wtd_quantile
from .misc import renameModelMatrixColumns


def estimateBetaPriorVar(
    obj, betaPriorMethod="weighted", upperQuantile=0.05, modelMatrix=None
):
    """
    steps for estimating the beta prior variance

    This lower-level function is called within :func:`DESeq` or
    :func:`nbinomWaldTest`. End users should use those higher-level functions.

    Arguments
    ---------
    obj : DESeqDataSet
        a :class:`.DESeqDataSet`
    betaPriorMethod : { "weighted", "quantile" }
        the method for calculating the beta prior variance:

        - :code:`"quantile"` matches a normal distribution using the upper
          quantile of the finite MLE betas.
        - :code:`"weighted"` matches a normal distribution using the upper
          quantile, but weighting by the variance of the MLE betas.

    upperQuantile : float
        the upper quantile to be used for the method of beta prior variance
        estimation
    modelMatrix : design matrix
        an optional design matrix, typically left :code:`None` and built within
        the function

    Returns
    -------
    ndarray
        the vector of variances for the prior on the beta in the :func:`DESeq`
        GLM
    """
    objNZ = obj[:, ~obj.var["allZero"]]

    betaMatrix = objNZ.var.filter(regex="MLE_")
    colnamesBM = [s.replace("MLE_", "") for s in betaMatrix.columns]
    # renaming in reverse:
    # make these standard colnames as from patsy.dmatrix
    convertNames = renameModelMatrixColumns(obj.obs, obj.design)
    convertNames = {y: x for x, y in convertNames.items()}
    colnamesBM = pd.Index(
        [convertNames[x] if x in convertNames else x for x in colnamesBM]
    )
    betaMatrix.columns = colnamesBM

    # this is the model matrix from an MLE run
    if modelMatrix is None:
        modelMatrix = obj.design
    modelMatrixType = obj.modelMatrixType

    if betaPriorMethod not in ["weighted", "quantile"]:
        raise ValueError(f"invalid value for betaPriorMethod: {betaPriorMethod}")

    # estimate the variance of the prior on betas
    # if expanded, first calculate LFC for all possible contrasts
    if modelMatrixType == "expanded":
        betaMatrix = objNZ.addAllContrasts(betaMatrix)

    # weighting by 1/Var(log(K))
    # Var(log(K)) ~ Var(K)/mu**2 = 1/mu + alpha
    # and using the fitted alpha
    if "dispFit" in objNZ.var:
        dispFit = objNZ.var["dispFit"]
    else:
        # betaPrior routine could have been called w/o the dispersion fitted trend
        dispFit = np.mean(objNZ.var["dispersion"])

    varlogk = 1 / objNZ.var["baseMean"] + dispFit
    weights = 1 / varlogk

    if betaMatrix.shape[0] > 1:

        def _anon(x):
            # this test removes genes which have betas tending to +/- infinity
            useFinite = np.abs(x) < 10
            # if no more betas pass test, return wide prior
            if np.sum(useFinite) == 0:
                return 1e6
            else:
                if betaPriorMethod == "quantile":
                    return matchUpperQuantileForVariance(x[useFinite], upperQuantile)
                else:
                    return matchWeightedUpperQuantileForVariance(
                        x[useFinite], weights[useFinite], upperQuantile
                    )

        betaPriorVar = betaMatrix.apply(_anon, 0)

    else:
        betaPriorVar = betaMatrix**2

    # TODO superfluous, no?
    betaPriorVar.columns = betaMatrix.columns

    # intercept set to wide prior
    if "Intercept" in betaPriorVar.columns:
        betaPriorVar["Intercept"] = 1e6

    if modelMatrixType == "expanded":
        # bring over beta priors from the GLM fit without prior.
        # for factors: prior variance of each level are the average of the
        # prior variances for the levels present in the previous GLM fit
        betaPriorVar = objNZ.averagePriorsOverLevels(betaPriorVar)

    return betaPriorVar


def matchUpperQuantileForVariance(x, upperQuantile=0.05):
    sdEst = np.quantile(np.abs(x), 1 - upperQuantile) / scipy.stats.norm.ppf(
        1 - upperQuantile / 2
    )
    return sdEst**2


def matchWeightedUpperQuantileForVariance(x, weights, upperQuantile=0.05):
    sdEst = wtd_quantile(
        np.abs(x), weights=weights, probs=(1 - upperQuantile), normwt=True
    ) / scipy.stats.norm.ppf(1 - upperQuantile / 2)
    return sdEst**2



import unittest
import numpy as np

from inmoose.utils import rnbinom
from inmoose.edgepy import aveLogCPM, DGEList


class test_aveLogCPM(unittest.TestCase):
    def setUp(self):
        y = np.array(rnbinom(80, size=5, mu=20, seed=42)).reshape((20, 4))
        y = np.vstack(([0, 0, 0, 0], [0, 0, 2, 2], y))
        group = np.array([1, 1, 2, 2])
        self.d = DGEList(y, group=group)

    def test_aveLogCPM(self):
        ref = np.array(
            [
                12.12019,
                12.70436,
                15.55656,
                15.07438,
                15.53215,
                15.56454,
                15.77740,
                15.83572,
                15.29030,
                15.58778,
                15.77746,
                15.38536,
                15.78073,
                15.76788,
                15.74191,
                16.12822,
                15.31202,
                15.85351,
                16.13554,
                16.30177,
                15.82458,
                15.52800,
            ]
        )
        self.assertTrue(np.allclose(self.d.aveLogCPM(), ref, atol=1e-5, rtol=0))

        self.assertTrue(
            np.allclose(self.d.aveLogCPM(dispersion=np.nan), ref, atol=1e-5, rtol=0)
        )
        dispersion = [0.05 for i in range(22)]
        dispersion[0] = np.nan
        self.assertTrue(
            np.allclose(self.d.aveLogCPM(dispersion=dispersion), ref, atol=1e-5, rtol=0)
        )

        self.d.samples.lib_size = None
        self.assertTrue(np.allclose(self.d.aveLogCPM(), ref, atol=1e-5, rtol=0))

        # Check special cases
        self.assertEqual(aveLogCPM(np.full((0, 0), 42)), 0)
        self.assertTrue(
            np.allclose(
                aveLogCPM(np.zeros(shape=(3, 2))),
                [18.34661, 18.34661, 18.34661],
                atol=1e-5,
                rtol=0,
            )
        )
        with self.assertRaisesRegex(ValueError, expected_regex="y should be a matrix"):
            aveLogCPM(42)



import unittest
import numpy as np

from inmoose.deseq2 import makeExampleDESeqDataSet, collapseReplicates


class Test(unittest.TestCase):
    def test_collapse(self):
        """test that collapse replicates works"""
        dds = makeExampleDESeqDataSet(n=10, m=8)
        dds.obs["sample"] = np.repeat([1, 2, 3, 4], 2)
        dds.obs["run"] = np.arange(8)
        dds2 = collapseReplicates(dds, groupby=dds.obs["sample"], run=dds.obs["run"])
        self.assertTrue(np.all(dds2.counts()[0, :] == np.sum(dds.counts()[0:2, :], 0)))
        self.assertEqual(dds2.obs["runsCollapsed"][0], "0,1")



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2023 A. Behdenna, A. Nordor, J. Haziza, A. Gema and M. Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import logging
import numpy as np
from functools import partial
import mpmath as mp
import pandas as pd

from .covariates import make_design_matrix


# aprior and bprior are useful to compute "hyper-prior values"
# -> prior parameters used to estimate the prior gamma distribution for multiplicative batch effect
# aprior - calculates empirical hyper-prior values


def compute_prior(prior, gamma_hat, mean_only):
    """[summary]

    Arguments:
        prior {char} -- 'a' or 'b' depending of the prior to be calculated
        gamma_hat {matrix} -- matrix of additive batch effect
        mean_only {bool} -- True iff mean_only selected

    Returns:
        float -- [the prior calculated (aprior or bprior)
    """
    if mean_only:
        return 1
    m = np.mean(gamma_hat)
    s2 = np.var(gamma_hat)
    if prior == "a":
        return (2 * s2 + m * m) / s2
    elif prior == "b":
        return (m * s2 + m * m * m) / s2


def postmean(g_bar, d_star, t2_n, t2_n_g_hat):
    """estimates additive batch effect

    Arguments:
        g_bar {matrix} -- additive batch effect
        d_star {matrix} -- multiplicative batch effect
        t2_n {matrix} --
        t2_n_g_hat {matrix} --

    Returns:
        matrix -- estimated additive batch effect
    """
    return np.divide(t2_n_g_hat + d_star * g_bar, np.asarray(t2_n + d_star))


def postvar(sum2, n, a, b):
    """estimates multiplicative batch effect

    Arguments:
        sum2 {vector} --
        n {[type]} --
        a {float} -- aprior
        b {float} -- bprior

    Returns:
        matrix -- estimated multiplicative batch effect
    """
    return np.divide((np.multiply(0.5, sum2) + b), (np.multiply(0.5, n) + a - 1))


def it_sol(sdat, g_hat, d_hat, g_bar, t2, a, b, conv=0.0001, exit_iteration=10e5):
    """iterative solution for Empirical Bayesian method

    Arguments:
        sdat {matrix} --
        g_hat {matrix} -- average additive batch effect
        d_hat {matrix} -- average multiplicative batch effect
        g_bar {matrix} -- additive batch effect
        t2 {matrix} --
        a {float} -- aprior
        b {float} -- bprior

    Keyword Arguments:
        conv {float} -- convergence criterion (default: {0.0001})
        exit_iteration {float} -- maximum number of iterations before exit (default: {10e5})

    Returns:
        array list -- estimated additive and multiplicative batch effect
    """

    n = [len(i) for i in np.asarray(sdat)]
    t2_n = np.multiply(t2, n)
    t2_n_g_hat = np.multiply(t2_n, g_hat)
    g_old = np.ndarray.copy(g_hat)
    d_old = np.ndarray.copy(d_hat)
    change = 1
    count = 0  # number of steps needed (for diagnostic only)
    # convergence criteria, if new-old < conv, then stop
    while (change > conv) and (count < exit_iteration):
        g_new = postmean(
            g_bar, d_old, t2_n, t2_n_g_hat
        )  # updated additive batch effect
        sum2 = np.sum(
            np.asarray(
                np.square(
                    sdat - np.outer(g_new[0][0], np.ones(np.ma.size(sdat, axis=1)))
                )
            ),
            axis=1,
        )
        d_new = postvar(sum2, n, a, b)  # updated multiplicative batch effect
        change = max(
            np.amax(np.absolute(g_new - np.asarray(g_old)) / np.asarray(g_old)),
            np.amax(np.absolute(d_new - d_old) / d_old),
        )  # maximum difference between new and old estimate
        g_old = np.ndarray.copy(g_new)  # save value for g
        d_old = np.ndarray.copy(d_new)  # save value for d
        count += 1
    return (g_new, d_new)


# int_eprior - Monte Carlo integration function to find nonparametric adjustments
# Johnson et al (Biostatistics 2007, supp.mat.) show that we can estimate the multiplicative and additive batch effects with an integral
# This integral is numerically computed through Monte Carlo inegration (iterative method)


def int_eprior(sdat, g_hat, d_hat, precision):
    """int_eprior - Monte Carlo integration function to find nonparametric adjustments
        Johnson et al (Biostatistics 2007, supp.mat.) show that we can estimate the multiplicative and additive batch effects with an integral
        This integral is numerically computed through Monte Carlo inegration (iterative method)

    Arguments:
        sdat {matrix} -- data matrix
        g_hat {matrix} -- average additive batch effect
        d_hat {matrix} -- average multiplicative batch effect
        precision {float} -- level of precision for precision computing

    Returns:
        array list -- estimated additive and multiplicative batch effect
    """
    g_star = []
    d_star = []
    # use this variable to only print error message once if approximation used
    test_approximation = 0
    for i in range(len(sdat)):
        # additive batch effect
        g = np.asarray(np.delete(np.transpose(g_hat), i))
        # multiplicative batch effect
        d = np.asarray(np.delete(np.transpose(d_hat), i))
        x = np.asarray(np.transpose(sdat[i]))
        n = len(x)
        j = [1] * n
        dat = np.repeat(x, len(np.transpose(g)), axis=1)
        resid2 = np.square(dat - g)
        sum2 = np.dot(np.transpose(resid2), j)
        # /begin{handling high precision computing}
        temp_2d = 2 * d
        if precision == None:
            LH = np.power(1 / (np.pi * temp_2d), n / 2) * np.exp(
                np.negative(sum2) / (temp_2d)
            )

        else:  # only if precision parameter informed
            # increase the precision of the computing (if negative exponential too close to 0)
            mp.dps = precision
            buf_exp = np.array(list(map(mp.exp, np.negative(sum2) / (temp_2d))))
            buf_pow = np.array(
                list(map(partial(mp.power, y=n / 2), 1 / (np.pi * temp_2d)))
            )
            # print(buf_exp.dtype, buf_pow.dtype)
            LH = buf_pow * buf_exp  # likelihood
        # /end{handling high precision computing}
        LH = np.nan_to_num(LH)  # corrects NaNs in likelihood
        if np.sum(LH) == 0 and test_approximation == 0:
            test_approximation = 1  # this message won't appear again
            logging.info(
                "###\nValues too small, approximation applied to avoid division by 0.\nPrecision mode can correct this problem, but increases computation time.\n###"
            )

        if np.sum(LH) == 0:  # correction for LH full of 0.0
            LH[LH == 0] = np.exp(-745)
            g_star.append(np.sum(g * LH) / np.sum(LH))
            d_star.append(np.sum(d * LH) / np.sum(LH))
        else:
            g_star.append(np.sum(g * LH) / np.sum(LH))
            d_star.append(np.sum(d * LH) / np.sum(LH))
    adjust = np.asarray([np.asarray(g_star), np.asarray(d_star)])
    return adjust


def param_fun(
    i, s_data, batches, mean_only, gamma_hat, gamma_bar, delta_hat, t2, a_prior, b_prior
):
    """parametric estimation of batch effects

    Arguments:
        i {int} -- column index
        s_data {matrix} --
        batches {list list} -- list of list of batches' elements
        mean_only {bool} -- True iff mean_only selected
        gamma_hat {matrix} -- average additive batch effect
        gamma_bar {matrix} -- estimated additive batch effect
        delta_hat {matrix} -- average multiplicative batch effect
        t2 {matrix} --
        a_prior {float} -- aprior
        b_prior {float} -- bprior

    Returns:
        array list -- estimated adjusted additive and multiplicative batch effect
    """
    if (
        mean_only
    ):  # if mean_only, no need for complex method: batch effect is immediately calculated
        t2_n = np.multiply(t2[i], 1)
        t2_n_g_hat = np.multiply(t2_n, gamma_hat[i])
        gamma_star = postmean(
            gamma_bar[i], 1, t2_n, t2_n_g_hat
        )  # additive batch effect
        delta_star = [1] * len(s_data)  # multiplicative batch effect
    else:  # if not(mean_only) then use it_solve
        # additive and multiplicative batch effects
        gamma_star, delta_star = it_sol(
            np.transpose(np.transpose(s_data)[batches[i]]),
            gamma_hat[i],
            delta_hat[i],
            gamma_bar[i],
            t2[i],
            a_prior[i],
            b_prior[i],
        )
    return [gamma_star, delta_star]


def nonparam_fun(i, mean_only, delta_hat, s_data, batches, gamma_hat, precision):
    """non-parametric estimation

    Arguments:
        i {int} -- column index
        mean_only {bool} -- True iff mean_only selected
        delta_hat {matrix} -- estimated multiplicative batch effect
        s_data {matrix} --
        batches {list list} -- list of list of batches' elements
        gamma_hat {matrix} -- estimated additive batch effect
        precision {float} -- level of precision for precision computing

    Returns:
        array list -- estimated adjusted additive and multiplicative batch effect
    """
    if mean_only:  # if mean only, change delta_hat to vector of 1s
        delta_hat[i] = [1] * len(delta_hat[i])
    # use int_eprior for non-parametric estimation
    temp = int_eprior(
        np.transpose(np.transpose(s_data)[batches[i]]),
        gamma_hat[i],
        delta_hat[i],
        precision,
    )
    return [temp[0], temp[1]]


############
# pyComBat #
############


def check_mean_only(mean_only):
    """checks mean_only option

    Arguments:
        mean_only {boolean} -- user's choice about mean_only

    Returns:
        ()
    """
    if mean_only == True:
        logging.info("Using mean only version")


def calculate_mean_var(design, batches, ref, dat, n_batches, n_batch, n_array):
    """calculates the Normalisation factors

    Arguments:
        design {matrix} -- model matrix for all covariates
        batches {int list} -- list of unique batches
        ref {int} -- reference batch index
        dat {matrix} -- data matrix
        n_batches {int list} -- list of batches lengths
        n_array {int} -- total size of dataset

    Returns:
        B_hat {matrix} -- regression coefficients corresponding to the design matrix
        grand_mean {matrix} -- Mean for each gene and each batch
        var_pooled {matrix} -- Variance for each gene and each batch
    """
    logging.info("Standardizing Data across genes.")
    # B_hat is the vector of regression coefficients corresponding to the design matrix
    B_hat = np.linalg.solve(
        np.dot(design, np.transpose(design)), np.dot(design, np.transpose(dat))
    )

    # Calculates the general mean
    if ref is not None:
        grand_mean = np.transpose(B_hat[ref])
    else:
        grand_mean = np.dot(
            np.transpose([i / n_array for i in n_batches]), B_hat[0:n_batch]
        )
    # Calculates the general variance
    if ref is not None:  # depending on ref batch
        ref_dat = np.transpose(np.transpose(dat)[batches[ref]])
        var_pooled = np.dot(
            np.square(
                ref_dat
                - np.transpose(np.dot(np.transpose(design)[batches[ref]], B_hat))
            ),
            [1 / n_batches[ref]] * n_batches[ref],
        )
    else:
        var_pooled = np.dot(
            np.square(dat - np.transpose(np.dot(np.transpose(design), B_hat))),
            [1 / n_array] * n_array,
        )

    return (B_hat, grand_mean, var_pooled)


def calculate_stand_mean(grand_mean, n_array, design, n_batch, B_hat):
    """transform the format of the mean for substraction

    Arguments:
        grand_mean {matrix} -- Mean for each gene and each batch
        n_array {int} -- total size of dataset
        design {[type]} -- design matrix for all covariates including batch
        n_batch {int} -- number of batches
        B_hat {matrix} -- regression coefficients corresponding to the design matrix

    Returns:
        stand_mean {matrix} -- standardised mean
    """
    stand_mean = np.dot(np.transpose(np.mat(grand_mean)), np.mat([1] * n_array))
    # corrects the mean with design matrix information
    if design is not None:
        tmp = np.ndarray.copy(design)
        tmp[0:n_batch] = 0
        stand_mean = stand_mean + np.transpose(np.dot(np.transpose(tmp), B_hat))
    return stand_mean


def standardise_data(dat, stand_mean, var_pooled, n_array):
    """standardise the data: substract mean and divide by variance

    Arguments:
        dat {matrix} -- data matrix
        stand_mean {matrix} -- standardised mean
        var_pooled {matrix} -- Variance for each gene and each batch
        n_array {int} -- total size of dataset

    Returns:
        s_data {matrix} -- standardised data matrix
    """
    s_data = (dat - stand_mean) / np.dot(
        np.transpose(np.mat(np.sqrt(var_pooled))), np.mat([1] * n_array)
    )
    return s_data


def fit_model(design, n_batch, s_data, batches, mean_only, par_prior, precision, ref):
    logging.info("Fitting L/S model and finding priors.")

    # fraction of design matrix related to batches
    batch_design = design[0:n_batch]

    # gamma_hat is the vector of additive batch effect
    gamma_hat = np.linalg.solve(
        np.dot(batch_design, np.transpose(batch_design)),
        np.dot(batch_design, np.transpose(s_data)),
    )

    delta_hat = []  # delta_hat is the vector of estimated multiplicative batch effect

    if mean_only:
        # no variance if mean_only == True
        delta_hat = [np.asarray([1] * len(s_data))] * len(batches)
    else:
        for i in batches:  # feed incrementally delta_hat
            list_map = np.transpose(np.transpose(s_data)[i]).var(
                axis=1
            )  # variance for each row
            delta_hat.append(np.squeeze(np.asarray(list_map)))

    gamma_bar = list(map(np.mean, gamma_hat))  # vector of means for gamma_hat
    t2 = list(map(np.var, gamma_hat))  # vector of variances for gamma_hat

    # calculates hyper priors for gamma (additive batch effect)
    a_prior = list(map(partial(compute_prior, "a", mean_only=mean_only), delta_hat))
    b_prior = list(map(partial(compute_prior, "b", mean_only=mean_only), delta_hat))

    # initialise gamma and delta for parameters estimation
    gamma_star = np.empty((n_batch, len(s_data)))
    delta_star = np.empty((n_batch, len(s_data)))

    if par_prior:
        # use param_fun function for parametric adjustments (cf. function definition)
        logging.info("Finding parametric adjustments.")
        results = list(
            map(
                partial(
                    param_fun,
                    s_data=s_data,
                    batches=batches,
                    mean_only=mean_only,
                    gamma_hat=gamma_hat,
                    gamma_bar=gamma_bar,
                    delta_hat=delta_hat,
                    t2=t2,
                    a_prior=a_prior,
                    b_prior=b_prior,
                ),
                range(n_batch),
            )
        )
    else:
        # use nonparam_fun for non-parametric adjustments (cf. function definition)
        logging.info("Finding nonparametric adjustments")
        results = list(
            map(
                partial(
                    nonparam_fun,
                    mean_only=mean_only,
                    delta_hat=delta_hat,
                    s_data=s_data,
                    batches=batches,
                    gamma_hat=gamma_hat,
                    precision=precision,
                ),
                range(n_batch),
            )
        )

    for i in range(n_batch):  # store the results in gamma/delta_star
        results_i = results[i]
        gamma_star[i], delta_star[i] = results_i[0], results_i[1]

    # update if reference batch (the reference batch is not supposed to be modified)
    if ref is not None:
        len_gamma_star_ref = len(gamma_star[ref])
        gamma_star[ref] = [0] * len_gamma_star_ref
        delta_star[ref] = [1] * len_gamma_star_ref

    return (gamma_star, delta_star, batch_design)


def adjust_data(
    s_data,
    gamma_star,
    delta_star,
    batch_design,
    n_batches,
    var_pooled,
    stand_mean,
    n_array,
    ref,
    batches,
    dat,
):
    """Adjust the data -- corrects for estimated batch effects

    Arguments:
        s_data {matrix} -- standardised data matrix
        gamma_star {matrix} -- estimated additive batch effect
        delta_star {matrix} -- estimated multiplicative batch effect
        batch_design {matrix} -- information about batches in design matrix
        n_batches {int list} -- list of batches lengths
        stand_mean {matrix} -- standardised mean
        var_pooled {matrix} -- Variance for each gene and each batch
        n_array {int} -- total size of dataset
        ref {int} -- the index of the reference batch in the batch list
        batches {int list} -- list of unique batches
        dat

    Returns:
        bayes_data [matrix] -- data adjusted for correction of batch effects
    """
    # Now we adjust the data:
    # 1. substract additive batch effect (gamma_star)
    # 2. divide by multiplicative batch effect (delta_star)
    logging.info("Adjusting the Data")
    bayes_data = np.transpose(s_data)
    j = 0
    for i in batches:  # for each batch, specific correction
        bayes_data[i] = (
            bayes_data[i] - np.dot(np.transpose(batch_design)[i], gamma_star)
        ) / np.transpose(
            np.outer(np.sqrt(delta_star[j]), np.asarray([1] * n_batches[j]))
        )
        j += 1

    # renormalise the data after correction:
    # 1. multiply by variance
    # 2. add mean
    bayes_data = (
        np.multiply(
            np.transpose(bayes_data),
            np.outer(np.sqrt(var_pooled), np.asarray([1] * n_array)),
        )
        + stand_mean
    )

    # correction for reference batch
    if ref is not None:
        bayes_data[batches[ref]] = dat[batches[ref]]

    # returns the data corrected for batch effects
    return bayes_data


def pycombat_norm(
    counts,
    batch,
    covar_mod=None,
    par_prior=True,
    prior_plots=False,
    mean_only=False,
    ref_batch=None,
    precision=None,
    na_cov_action="raise",
    **kwargs,
):
    """Corrects batch effect in microarray expression data. Takes an gene expression file and a list of known batches corresponding to each sample.

    Arguments
    ---------
    counts : matrix
        expression matrix (dataframe or numpy array). It contains the information about the gene expression (rows) for each sample (columns).
    batch : list
        batch indices. Must have as many elements as the number of columns in the expression matrix.
    covar_mod : list or matrix, optional
        model matrix (dataframe, list or numpy array) for one or multiple covariates to include in linear model (signal
        from these variables are kept in data after adjustment). Covariates have to be categorial,
        they can not be continious values (default: `None`).
    par_prior : bool, optional
        False for non-parametric estimation of batch effects (default: `True`).
    prior_plots : bool, optional
        True if requires to plot the priors (default: `False`). -- Not implemented yet!
    mean_only : bool, optional
        True iff just adjusting the means and not individual batch effects (default: `False`).
    ref_batch, optional
        batch id of the batch to use as reference (default: `None`)
    precision : float, optional
        level of precision for precision computing (default: `None`).
    na_cov_action : str
        Option to choose the way to handle missing covariates

        - :code:`"raise"` raise an error if missing covariates and stop the code
        - :code:`"remove"` remove samples with missing covariates and raise a warning
        - :code:`"fill"` handle missing covariates, by creating a distinct
          covariate per batch
        (default: :code:`"raise"`)

    Returns
    -------
    matrix
        the input expression matrix adjusted for batch effects.
        same type as the input `data`
    """
    check_mean_only(mean_only)

    # Handle batches, covariates and prepare design matrix
    vci = make_design_matrix(
        counts, batch, covar_mod, ref_batch, na_cov_action=na_cov_action
    )

    dat = vci.counts
    list_samples = vci.list_samples
    list_genes = vci.list_genes
    batch = vci.batch
    design = np.transpose(vci.design)
    ref = vci.ref_batch_idx

    n_batch = vci.n_batch
    batches_ind = [vci.batch_composition[b] for b in batch.categories]
    batch_sizes = [len(v) for v in batches_ind]

    n_sample = dat.shape[1]

    B_hat, grand_mean, var_pooled = calculate_mean_var(
        design, batches_ind, ref, dat, batch_sizes, n_batch, n_sample
    )
    stand_mean = calculate_stand_mean(grand_mean, n_sample, design, n_batch, B_hat)
    s_data = standardise_data(dat, stand_mean, var_pooled, n_sample)
    gamma_star, delta_star, batch_design = fit_model(
        design, n_batch, s_data, batches_ind, mean_only, par_prior, precision, ref
    )
    bayes_data = adjust_data(
        s_data,
        gamma_star,
        delta_star,
        batch_design,
        batch_sizes,
        var_pooled,
        stand_mean,
        n_sample,
        ref,
        batches_ind,
        dat,
    )

    if isinstance(counts, pd.DataFrame):
        return pd.DataFrame(bayes_data, columns=list_samples, index=list_genes)
    else:
        return bayes_data



from .core import DESeq
from .DESeqDataSet import DESeqDataSet, makeExampleDESeqDataSet
from .DESeqTransform import DESeqTransform
from .dispersions import (
    estimateDispersionsGeneEst,
    estimateDispersionsFit,
    estimateDispersionsMAP,
    estimateDispersionsPriorVar,
)
from .estimateSizeFactors import estimateSizeFactorsForMatrix
from .lrt import nbinomLRT
from .outliers import replaceOutliers
from .parallel import estimateMLEForBetaPriorVar
from .prior import estimateBetaPriorVar
from .replicates import collapseReplicates
from .vst import varianceStabilizingTransformation
from .wald import nbinomWaldTest



import unittest

from inmoose.sim.splat import sim_rnaseq

N = 1000
M = 10000


class Test(unittest.TestCase):
    def test_sim_rnaseq(self):
        """
        check that RNA-Seq data simulation works properly
        """
        data = sim_rnaseq(N, M, random_state=42)
        self.assertEqual(data.shape, (N, M))

    def test_sim_scrnaseq(self):
        """
        check that scRNA-Seq data simulation works properly
        """
        data = sim_rnaseq(N, M, single_cell=True, random_state=51)
        self.assertEqual(data.shape, (N, M))



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/core.R' of the Bioconductor DESeq2 package
# (version 3.16).


import numpy as np

from ..utils import Factor


def collapseReplicates(obj, groupby, run, renameCols=True):
    """
    Collapse technical replicates in an AnnData or DESeqDataSet

    Collapse the samples (rows) in `obj` by summing within levels
    of a grouping factor `groupby`. The purpose of this function
    if to sum up read counts from technical replicates to create an object
    with a single row of read counts for each sample.
    Note: by "technical replicates", we mean multiple sequencing runs of the same
    library, in contrast to "biological replicates" in which multiple
    libraries are prepared from separate biological units.
    Optionally renames the columns of returned object with the levels of the
    grouping factor.

    Arguments
    ---------
    obj : AnnData or DESeqDataSet
    groupby : Factor
        a grouping factor, as long as the rows of obj
    run : ??, optional
        the names of each unique row in obj
        if provided, a new row "runsCollapsed" will be added to obj.obs
        which pastes together the names of `run`
    renameCols : bool, optional
        whether to rename the rows of the returned object using the levels
        of the grouping factor

    Returns
    -------
    AnnData or DESeqDataSet
        an object with as many rows as levels in `groupby`.
        This object has count data which is summed from the various rows which
        are grouped together, and its .obs is subset using the first row
        for each group in `groupby`.

    Examples
    --------
    >>> dds = makeExampleDESeqDataSet(m=12)
    >>> # make data with two technical replicates for three samples
    >>> dds.obs["sample"] = Factor(np.repeat(np.arange(1,10), [2,1,1,2,1,1,2,1,1]))
    >>> dds.obs["run"] = [f"run{i}" for i in range(12)]
    >>> ddsColl = collapseReplicates(dds, dds.obs["sample"], dds.obs["run"])
    >>> # examine the clinical data and rows names of the collapsed data
    >>> ddsColl.obs
    ???
    >>> ddsColl.index
    ???
    >>> # check that the sum of the counts for "sample0" is the same
    ... # as the counts in the "sample0" row in ddsColl
    >>> matchFirstLevel = dds.obs["sample"] == dds.obs["sample"].categories[0]
    >>> np.all(np.sum(dds[matchFirstLevel,:].counts(), axis=0) == ddsColl[0,:].counts())
    True
    """
    if not isinstance(groupby, Factor):
        groupby = Factor(groupby)
    groupby = groupby.droplevels()
    if len(groupby) != obj.n_obs:
        raise ValueError("groupby should have as many elements as observations in obj")

    countdata = [
        np.sum(obj.counts()[groupby == c, :], axis=0) for c in groupby.categories
    ]
    rowsToKeep = [np.nonzero(groupby == c)[0][0] for c in groupby.categories]
    collapsed = obj[rowsToKeep, :]
    collapsed.X = countdata
    if run is not None:
        if len(groupby) != len(run):
            raise ValueError("groupby and run should have the same length")
        collapsed.obs["runsCollapsed"] = [
            ",".join([f"{e}" for e in run[groupby == c]]) for c in groupby.categories
        ]

    if renameCols:
        collapsed.index = groupby.categories

    assert np.sum(obj.counts()) == np.sum(collapsed.counts())

    return collapsed



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/core.R' of the Bioconductor DESeq2 package
# (version 3.16).


import numpy as np


def nbinomLRT(
    obj,
    full=None,
    reduced=None,
    betaTol=1e-8,
    maxit=100,
    useOptim=True,
    quiet=False,
    useQR=True,
    minmu=None,
    type_="DESeq2",
):
    """
    Likelihood ratio test (chi-squared test) for GLMs

    This function tests for significance of change in deviance between a full
    and reduced model which are provided as :code:`formula`. Fitting uses
    previously calculated :attr:`.DESeqDataSet.sizeFactors` (or
    :attr:`.DESeqDataSet.normalizationFactors`) and dispersion estimates.

    The difference in deviance is compared to chi-squared distribution with df
    = (reduced residual degrees of freedom - full residual degrees of freedom).
    This function is an alternative to :func:`nbinomWaldTest`.

    See Also
    --------
    DESeq
    nbinomWaldTest

    Arguments
    ---------
    obj : DESeqDataSet
        a DESeqDataSet
    full
        the full model formula, this should be the formula in
        :code:`obj.design`.  Alternatively can be a matrix.
    reduced
        a reduced formula to compare against, e.g. the full model with a term
        or terms of interest removed.
    betaTol : float
        control parameter defining convergence
    maxit : int
        the maximum number of iterations to allow for convergence of the
        coefficient vector
    useOptim : bool
        whether to use the native optim function on columns which do not
        converge within :code:`maxit` iterations
    quiet : bool
        whether to print messages at each step
    useQR : bool
        whether to use the QR decomposition of the design matrix while fitting
        the GLM
    minmu : float
        lower bound on the estimated count while fitting the GLM
    type_ : "DESeq2" or "glmGamPoi"
        If :code:`"DESeq2"`, a classical Likelihood ratio test based on the Chi-squared distribution is conducted.

        If :code:`"glmGamPoi"` and previously the dispersion has been estimated
        with :code:`"glmGamPoi"` as well, a quasi-likelihood ratio test based
        on the F-distribution is conducted. It is supposed to be more accurate,
        because it takes the uncertainty of dispersion estimate into account
        in the same way that a t-test improves upon a Z-test.

    Returns
    -------
    DESeqDataSet
        the input :code:`obj` with new results columns accessible through
        :meth:`.DESeqDataSet.results`. The coefficients and standard errors are
        reported on a log2 scale.
    """

    if type_ not in ["DESeq2", "glmGamPoi"]:
        raise ValueError(f"invalid value for type_: {type_}")
    if "dispersion" not in obj.var:
        raise ValueError(
            "testing requires dispersion estimates, first call estimateDispersions()"
        )

    if reduced is None:
        raise ValueError(
            'provide a reduced formula for the LRT, e.g. nbinomLRT(obj, reduced="~1")'
        )

    # run check on the formula
    raise NotImplementedError()


def checkLRT(full, reduced):
    """check for LRT formulas, written as function to share code between DESeq and nbinomLRT"""
    reducedNotInFull = ~np.isin(reduced.design_info.terms, full.design_info.terms)
    if np.any(reducedNotInFull):
        reducedVars = [
            v for (v, x) in zip(reduced.design_info.term_names, reducedNotInFull) if x
        ]
        raise ValueError(
            f"the following variables in the reduced formula not in the full formula: {' '.join(reducedVars)}"
        )

    full.design_info.terms



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/glmfit.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np

from ..utils import asfactor
from .DGEGLM import DGEGLM
from .makeCompressedMatrix import _compressDispersions, _compressOffsets
from .mglmLevenberg import mglmLevenberg
from .mglmOneWay import designAsFactor, mglmOneWay
from .nbinomDeviance import nbinomDeviance
from .predFC import predFC

from patsy import dmatrix


def glmFit_DGEList(self, design=None, dispersion=None, prior_count=0.125, start=None):
    """
    Fit a negative binomial generalized log-linear model to the read counts for
    each gene. Conduct genewise statistical tests for a given coefficient or
    coefficient contrast.

    See also
    --------
    glmFit

    Arguments
    ---------
    design : matrix, optional
        design matrix for the genewise linear models. Must be of full column
        rank. Defaults to a single column of ones, equivalent to treating the
        columns as replicate libraries.
    dispersion : float or array_like
        scalar, vector or matrix of negative binomial dispersions. Can be a
        common value for all genese, a vector of dispersion values with one for
        each gene, or a matrix of dispersion values with one for each observation.
        If :code:`None`, it will be extracted from :code:`y`, with order of
        precedence: genewise dispersion, trended dispersion, common dispersion.
    prior_count : float
        average prior count to be added to observation to shrink the estimated
        log-fold-change towards zero.
    start : matrix, optional
        initial estimates for the linear model coefficients

    Returns
    -------
    DGEList
        object containing the data about the fit
    """

    # The design matrix defaults to the oneway layout defined by self.samples.group
    # If there is only one group, then the design matrix is left None so that a matrix with a single intercept column will be set later by glmFit.
    if design is None:
        design = self.design
        if design is None:
            group = asfactor(self.samples.group).droplevels()
            if group.nlevels() > 1:
                design = dmatrix("~C(self.samples.group)")

    if dispersion is None:
        dispersion = self.getDispersion()
    if dispersion is None:
        raise ValueError("No dispersion values found in DGEList object")
    offset = self.getOffset()
    if self.AveLogCPM is None:
        self.AveLogCPM = self.aveLogCPM()

    fit = glmFit(
        y=self.counts,
        design=design,
        dispersion=dispersion,
        offset=offset,
        lib_size=None,
        weights=self.weights,
        prior_count=prior_count,
        start=start,
    )

    fit.samples = self.samples
    fit.genes = self.genes
    fit.prior_df = self.prior_df
    fit.AveLogCPM = self.AveLogCPM
    return fit


def glmFit(
    y,
    design=None,
    dispersion=None,
    offset=None,
    lib_size=None,
    weights=None,
    prior_count=0.125,
    start=None,
):
    """
    Fit a negative binomial generalized log-linear model to the read counts for
    each gene. Conduct genewise statistical tests for a given coefficient or
    coefficient contrast.

    This function implements one of the GLM methods developed by McCarthy et al.
    (2012) [1]_.

    :code:`glmFit` fits genewise negative binomial GLMs, all with the same
    design matrix but possibly different dispersions, offsets and weights.
    When the design matrix defines a one-way layout, or can be re-parameterized
    to a one-way layout, the GLMs are fitting very quickly using
    :func:`mglmOneGroup`. Otherwise the default fitting method, implemented in
    :func:`mglmLevenberg`, uses a Fisher scoring algorithm with Levenberg-style
    damping.

    Positive :code:`prior_count` cause the returned coefficients to be shrunk in
    such a way that fold-changes between the treatment conditions are decreased.
    In particular, infinite fold-changes are avoided. Larger values cause more
    shrinkage. The returned coefficients are affected but not the likelihood
    ratio tests or p-values.

    See also
    --------
    mglmOneGroup : low-level computations
    mglmLevenberg : low-level computations

    Arguments
    ---------
    y : matrix
        matrix of counts
    design : matrix, optional
        design matrix for the genewise linear models. Must be of full column
        rank. Defaults to a single column of ones, equivalent to treating the
        columns as replicate libraries.
    dispersion : float or array_like
        scalar, vector or matrix of negative binomial dispersions. Can be a
        common value for all genese, a vector of dispersion values with one for
        each gene, or a matrix of dispersion values with one for each
        observation.
    offset : float or array_like, optional
        matrix of the same shape as :code:`y` giving offsets for the log-linear
        models. Can be a scalar or a vector of length :code:`y.shape[1]`, in
        which case it is broadcasted to the shape of :code:`y`.
    lib_size : array_like, optional
        vector of length :code:`y.shape[1]` giving library sizes. Only used if
        :code:`offset=None`, in which case :code:`offset` is set to
        :code:`log(lib_size)`. Defaults to :code:`colSums(y)`.
    weights : matrix, optional
        prior weights for the observations (for each library and gene) to be
        used in the GLM calculations
    prior_count : float
        average prior count to be added to observation to shrink the estimated
        log-fold-change towards zero.
    start : matrix, optional
        initial estimates for the linear model coefficients

    Returns
    -------
    DGEGLM
        object containing:

        - :code:`counts`, the input matrix of counts
        - :code:`design`, the input design matrix
        - :code:`weights`, the input weights matrix
        - :code:`offset`, matrix of linear model offsets
        - :code:`dispersion`, vector of dispersions used for the fit
        - :code:`coefficients`, matrix of estimated coefficients from the GLM
          fits, on the natural log scale, of size :code:`y.shape[0]` by
          :code:`design.shape[1]`.
        - :code:`unshrunk_coefficients`, matrix of estimated coefficients from
          the GLM fits when no log-fold-changes shrinkage is applied, on the
          natural log scale, of size :code:`y.shape[0]` by
          :code:`design.shape[1]`. It exists only when :code:`prior_count` is
          not 0.
        - :code:`fitted_values`, matrix of fitted values from GLM fits, same
          shape as :code:`y`
        - :code:`deviance`, numeric vector of deviances, one for each gene

    References
    ----------
    .. [1] D. J. McCarthy, Y. Chen, G. K. Smyth. 2012. Differential expression
       analysis of multifactor RNA-Seq experiments with respect to biological
       variation. Nucleic Acids Research 40, 4288-4297. :doi:`10.1093/nar/gks042`
    """
    # Check y
    y = np.asarray(y, order="F")
    (ntag, nlib) = y.shape

    # Check design
    if design is None:
        design = np.ones(shape=(nlib, 1), order="F")
    else:
        design = np.asarray(design, order="F")
        if design.shape[0] != nlib:
            raise ValueError("design should have as many rows as y has columns")
        if np.linalg.matrix_rank(design) < design.shape[1]:
            raise ValueError(
                "Design matrix is not full rank. Some coefficients are not estimable"
            )

    # Check dispersion
    if dispersion is None:
        raise ValueError("No dispersion values provided")
    dispersion = np.asanyarray(dispersion, order="F")
    # TODO check dispersion for NaN and non-numeric values
    if dispersion.shape not in [(), (1,), (ntag,), y.shape]:
        raise ValueError("Dimensions of dispersion do not agree with dimensions of y")
    dispersion_mat = _compressDispersions(y, dispersion)

    # Check offset
    if offset is not None:
        # TODO check that offset is numeric
        offset = np.asanyarray(offset, order="F")
        if offset.shape not in [(), (1,), (nlib,), y.shape]:
            raise ValueError("Dimensions of offset do not agree with dimensions of y")

    # Check lib_size
    if lib_size is not None:
        # TODO check that lib_size is numeric
        lib_size = np.asarray(lib_size, order="F")
        if lib_size.shape not in [(), (1,), (nlib,)]:
            raise ValueError("lib_size has wrong length, should agree with ncol(y)")

    # Consolidate lib_size and offset into a compressed matrix
    offset = _compressOffsets(y=y, lib_size=lib_size, offset=offset)

    # weights are checked in lower-level functions

    # Fit the tagwise GLMs
    # If the design is equivalent to a oneway layout, use a shortcut algorithm
    group = designAsFactor(design)
    if group.nlevels() == design.shape[1]:
        (coef, fitted_values) = mglmOneWay(
            y,
            design=design,
            group=group,
            dispersion=dispersion_mat,
            offset=offset,
            weights=weights,
            coef_start=start,
        )
        deviance = nbinomDeviance(
            y=y, mean=fitted_values, dispersion=dispersion_mat, weights=weights
        )
        fit_method = "oneway"
        fit = (coef, fitted_values, deviance, None, None)
    else:
        fit = mglmLevenberg(
            y,
            design=design,
            dispersion=dispersion_mat,
            offset=offset,
            weights=weights,
            coef_start=start,
            maxit=250,
        )
        fit_method = "levenberg"

    # Prepare output
    fit = DGEGLM(fit)
    fit.counts = y
    fit.method = fit_method
    if prior_count > 0:
        fit.unshrunk_coefficients = fit.coefficients
        fit.coefficients = predFC(
            y,
            design,
            offset=offset,
            dispersion=dispersion_mat,
            prior_count=prior_count,
            weights=weights,
        ) * np.log(2)

    # FIXME (from original R source) we are not allowing missing values, so df.residual must be same for all tags
    fit.df_residual = np.full(ntag, nlib - design.shape[1])
    fit.design = design
    fit.offset = offset
    fit.dispersion = dispersion
    fit.weights = weights
    fit.prior_count = prior_count
    return fit
